# Acknowledgements:
# Author: Robert H Cudmore
# Date: 20210603

import os, time, sys
import copy  # For copy.deepcopy() of bAnalysis
import uuid  # to generate unique key on bAnalysis spike detect
import pathlib  # ned to use this (introduced in Python 3.4) to maname paths on Windows, stop using os.path

from typing import List, Union, Optional

import numpy as np
import pandas as pd
import requests, io  # too load from the web

# for old code that was compressing hdf5 files
# from subprocess import call # to call ptrepack (might fail on windows???)
# from pprint import pprint
# import tables.scripts.ptrepack  # to save compressed .h5 file
# import shutil

import sanpy
import sanpy.h5Util

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)

# Turn off pandas save h5 performance warnnig
# see: https://github.com/pandas-dev/pandas/issues/3622
# /home/cudmore/Sites/SanPy/sanpy/analysisDir.py:478: PerformanceWarning:
# your performance may suffer as PyTables will pickle object types that it cannot
# map directly to c-types [inferred_type->mixed-integer,key->block0_values] [items->Index(['detectionClass', '_isAnalyzed', '_dataPointsPerMs', '_sweepList',
import warnings

warnings.filterwarnings("ignore", category=pd.io.pytables.PerformanceWarning)

_sanpyColumns = {
    #'Idx': {
    #    #'type': int,
    #    'type': float,
    #    'isEditable': False,
    # },
    "L": {
        # loaded
        "type": str,
        "isEditable": False,
    },
    "A": {
        # analyzed
        "type": str,
        "isEditable": False,
    },
    "S": {
        # saved
        "type": str,
        "isEditable": False,
    },
    "N": {
        # number of spikes
        "type": int,
        "isEditable": False,
    },
    "E": {
        # number of errors
        "type": int,
        "isEditable": False,
    },
    # 'I': {
    #     # include
    #     # problems with isinstance(bool), just using string
    #     'type': bool,
    #     'isEditable': True,
    # },
    "File": {
        "type": str,
        "isEditable": False,
    },
    "Dur(s)": {
        "type": float,
        "isEditable": False,
    },
    "Channels": {  # For Thianne
        "type": int,
        "isEditable": False,
    },
    "Sweeps": {
        "type": int,
        "isEditable": False,
    },
    "Epochs": {  # For Thianne
        "type": int,
        "isEditable": False,
    },
    "kHz": {
        "type": float,
        "isEditable": False,
    },
    "Mode": {
        "type": str,
        "isEditable": False,
    },
    "Start(s)": {
        "type": float,
        "isEditable": False,
    },
    "Stop(s)": {
        "type": float,
        "isEditable": False,
    },
    "dvdtThreshold": {
        "type": float,
        "isEditable": True,
    },
    "mvThreshold": {
        "type": float,
        "isEditable": True,
    },
    "Cell Type": {
        "type": str,
        "isEditable": True,
    },
    "Sex": {
        "type": str,
        "isEditable": True,
    },
    "Condition": {
        "type": str,
        "isEditable": True,
    },
    "Notes": {
        "type": str,
        "isEditable": True,
    },
    # kymograph interface
    "kLeft": {
        "type": int,
        "isEditable": False,
    },
    "kTop": {
        "type": int,
        "isEditable": False,
    },
    "kRight": {
        "type": int,
        "isEditable": False,
    },
    "kBottom": {
        "type": int,
        "isEditable": False,
    },
    "relPath": {
        "type": str,
        "isEditable": False,
    },
    "uuid": {
        "type": str,
        "isEditable": False,
    },
}


"""
Columns to use in display in file table (pyqt, dash, vue).
We require type so we can edit with QAbstractTableModel.
Critical for qt interface to allow easy editing of values while preserving type
"""


def _fixRelPath(folderPath, dfTable: pd.DataFrame, fileList: List[str]):
    """
    Was not assigning relPath on initial load (no hd5 file).

    We need a path relative to location of loaded folder.
    This allows a folder of files and analysis to be moved (to a different machine)
    """

    # print('fileList:', fileList)
    # pprint(dfTable[['File', 'relPath']])

    # if dfTable is None:
    #     logger.error('no dfTable')

    n = len(dfTable)

    logger.info(f"Checking path for {n} file(s)")

    for rowIdx in range(n):
        file = dfTable.loc[rowIdx, "File"]
        relPath = dfTable.loc[rowIdx, "relPath"]

        # 20220422 why was this here?
        # if relPath:
        #    continue

        # print(rowIdx, file, relPath)
        for filePath in fileList:
            if filePath.find(file) != -1:
                logger.info(
                    f"    _fixRelPath() file idx {rowIdx} file:{file} now has relPath:{filePath}"
                )
                dfTable.loc[rowIdx, "relPath"] = filePath


def old_h5_printKey(hdfPath):
    logger.info(f"hdfPath: {hdfPath} has keys:")
    with pd.HDFStore(hdfPath, mode="r") as store:
        for key in store.keys():
            logger.info(f"    {key}")


class bAnalysisDirWeb:
    """
    Load a directory of .abf from the web (for now from GitHub).

    Will etend this to Box, Dropbox, other?.
    """

    def __init__(self, cloudDict):
        """
        Args: cloudDict (dict): {
                    'owner': 'cudmore',
                    'repo_name': 'SanPy',
                    'path': 'data'
                    }
        """
        self._cloudDict = cloudDict
        self.loadFolder()

    def loadFolder(self):
        """Load using cloudDict"""

        """
        # use ['download_url'] to download abf file (no byte conversion)
        response[0] = {
            name : 171116sh_0018.abf
            path : data/171116sh_0018.abf
            sha : 5f3322b08d86458bf7ac8b5c12564933142ffd17
            size : 2047488
            url : https://api.github.com/repos/cudmore/SanPy/contents/data/171116sh_0018.abf?ref=master
            html_url : https://github.com/cudmore/SanPy/blob/master/data/171116sh_0018.abf
            git_url : https://api.github.com/repos/cudmore/SanPy/git/blobs/5f3322b08d86458bf7ac8b5c12564933142ffd17
            download_url : https://raw.githubusercontent.com/cudmore/SanPy/master/data/171116sh_0018.abf
            type : file
            _links : {'self': 'https://api.github.com/repos/cudmore/SanPy/contents/data/171116sh_0018.abf?ref=master', 'git': 'https://api.github.com/repos/cudmore/SanPy/git/blobs/5f3322b08d86458bf7ac8b5c12564933142ffd17', 'html': 'https://github.com/cudmore/SanPy/blob/master/data/171116sh_0018.abf'}
        }
        """

        owner = self._cloudDict["owner"]
        repo_name = self._cloudDict["repo_name"]
        path = self._cloudDict["path"]
        url = f"https://api.github.com/repos/{owner}/{repo_name}/contents/{path}"

        # response is a list of dict
        response = requests.get(url).json()
        # print('response:', type(response))

        for idx, item in enumerate(response):
            if not item["name"].endswith(".abf"):
                continue
            print(idx)
            # use item['git_url']
            for k, v in item.items():
                print("  ", k, ":", v)

        #
        # test load
        download_url = response[1]["download_url"]
        content = requests.get(download_url).content

        fileLikeObject = io.BytesIO(content)
        ba = sanpy.bAnalysis(byteStream=fileLikeObject)
        # print(ba._abf)
        # print(ba.api_getHeader())
        ba.spikeDetect()
        print(ba.numSpikes)


class analysisDir:
    """
    Class to manage a list of files loaded from a folder.
    """

    sanpyColumns = _sanpyColumns
    """Dict of dict of column names and bookkeeping info.
    """

    theseFileTypes = [".abf", ".atf", ".csv", ".tif"]
    """File types to load.
    """

    def __init__(
        self,
        path: str = None,
        myApp=None,
        autoLoad: bool = False,
        folderDepth: Optional[int] = None,
    ):
        """Load and manage a list of files in a folder path.

        Use this as the main pandasModel for file list myTableView.

        TODO: extend to link to folder in cloud (start with box and/or github)

        Args:
            path (str): Path to folder
            myApp (sanpy.interface.sanpy_app): Optional
            autoLoad (bool):
            folderDepth (int):
            #cloudDict (dict): To load from cloud, for now  just github

        Notes:
            - Some functions are so self can mimic a pandas dataframe used by pandasModel.
                (shape, loc, loc_setter, iloc, iLoc_setter, columns, append, drop, sort_values, copy)
        """
        self.path: str = path
        self.myApp = myApp  # used to signal on building initial db
        self.autoLoad = autoLoad  # not used

        self.folderDepth = folderDepth  # specify int

        self._isDirty = False

        self._poolDf = None
        """See pool_ functions"""

        # keys are full path to file, if from cloud, key is 'cloud/<filename>'
        # holds bAnalysisObjects
        # needs to be a list so we can have files more than one
        # self.fileList = [] #OrderedDict()

        # TODO: refactor, we are not using the csv parth of this, just the filename
        # name of database file created/loaded from folder path
        self.dbFile = "sanpy_recording_db.csv"

        self._df = None
        # if autoLoad:
        if 1:
            self._df = self.loadHdf()
            if self._df is None:
                self._df = self.loadFolder(loadData=autoLoad)  # only used if no h5 file
                self._updateLoadedAnalyzed()

        # self._df = self.loadFolder(loadData=autoLoad)

        #
        self._checkColumns()
        self._updateLoadedAnalyzed()

        """
        logger.warning('remember: temporary fix with _fixRelPath()\n')
        tmpFileList = self.getFileList()
        _fixRelPath(self.path, self._df, tmpFileList)
        """

    def __iter__(self):
        self._iterIdx = 0
        return self

    def __next__(self):
        if self._iterIdx < self.numFiles:
            x = self._df.loc[self._iterIdx]["_ba"]
            self._iterIdx += 1
            return x
        else:
            raise StopIteration

    def __str__(self):
        totalDurSec = self._df["Dur(s)"].sum()
        theStr = f"analysisDir Num Files: {len(self)} Total Dur(s): {totalDurSec}"
        return theStr

    @property
    def isDirty(self):
        return self._isDirty

    def __len__(self):
        return len(self._df)

    @property
    def numFiles(self):
        return len(self._df)

    @property
    def shape(self):
        """
        Can't just return shape of _df, columns (like 'ba') may have been added
        Number of columns is based on self.columns
        """
        # return self._df.shape
        numRows = self._df.shape[0]
        numCols = len(self.columns)
        return (numRows, numCols)

    @property
    def loc(self):
        """Mimic pandas df.loc[]"""
        return self._df.loc

    @loc.setter
    def loc_setter(self, rowIdx, colStr, value):
        self._df.loc[rowIdx, colStr] = value

    @property
    def iloc(self):
        # mimic pandas df.iloc[]
        return self._df.iloc

    @iloc.setter
    def iLoc_setter(self, rowIdx, colIdx, value):
        self._df.iloc[rowIdx, colIdx] = value
        self._isDirty = True

    @property
    def at(self):
        # mimic pandas df.at[]
        return self._df.at

    @at.setter
    def at_setter(self, rowIdx, colStr, value):
        self._df.at[rowIdx, colStr] = value
        self._isDirty = True

    """
    @property
    def iat(self):
        # mimic pandas df.iat[]
        return self._df.iat

    @iat.setter
    def iat_setter(self, rowIdx, colStr, value):
        self._df.iat[rowIdx, colStr] = value
        self._isDirty = True
    """

    @property
    def index(self):
        return self._df.index

    @property
    def columns(self):
        # return list of column names
        return list(self.sanpyColumns.keys())

    def copy(self):
        return self._df.copy()

    def sort_values(self, Ncol, order):
        logger.info(f"sorting by column {self.columns[Ncol]} with order:{order}")
        self._df = self._df.sort_values(self.columns[Ncol], ascending=not order)
        # print(self._df)

    @property
    def columnsDict(self):
        return self.sanpyColumns

    def columnIsEditable(self, colName):
        return self.sanpyColumns[colName]["isEditable"]

    def columnIsCheckBox(self, colName):
        """All bool columns are checkbox

        TODO: problems with using type=bool and isinstance(). Kust using str 'bool'
        """
        type = self.sanpyColumns[colName]["type"]
        # isBool = isinstance(type, bool)
        isBool = type == "bool"
        # logger.info(f'{colName} {type(type)}, type:{type} {isBool}')
        return isBool

    def getDataFrame(self):
        """Get the underlying pandas DataFrame."""
        return self._df

    @property
    def numFiles(self):
        """Get the number of files. same as len()."""
        return len(self._df)

    def copyToClipboard(self):
        """
        TODO: Is this used or is copy to clipboard in pandas model?
        """
        if self.getDataFrame() is not None:
            self.getDataFrame().to_clipboard(sep="\t", index=False)
            logger.info("Copied to clipboard")

    def old_saveDatabase(self):
        """save dbFile .csv and hdf .gzip"""
        dbPath = os.path.join(self.path, self.dbFile)
        if self.getDataFrame() is not None:
            #
            logger.info(f'Saving "{dbPath}"')
            self.getDataFrame().to_csv(dbPath, index=False)
            self._isDirty = False

            #
            """
            hdfFile = os.path.splitext(self.dbFile)[0] + '.h5'
            hdfPath = os.path.join(self.path, hdfFile)
            logger.info(f'Saving "{hdfPath}"')
            #hdfStore = pd.HDFStore(hdfPath)
            start = time.time()
            complevel = 9
            complib = 'blosc:blosclz'
            with pd.HDFStore(hdfPath, mode='w', complevel=complevel, complib=complib) as hdfStore:
                hdfStore['df'] = self.getDataFrame()  # save it
            stop = time.time()
            logger.info(f'Saving took {round(stop-start,2)} seconds')
            """

    def old_getFrozenPath(self):
        if getattr(sys, "frozen", False):
            # running in a bundle (frozen)
            myPath = sys._MEIPASS
        else:
            # running in a normal Python environment
            # myPath = os.path.dirname(os.path.abspath(__file__))
            myPath = pathlib.Path(__file__).parent.absolute()
        return myPath

    def old_rebuildHdf(self):
        #
        # rebuild the file to remove old changes and reduce size
        tmpHdfFile = os.path.splitext(self.dbFile)[0] + "_tmp.h5"
        # tmpHdfPath = os.path.join(self.path, tmpHdfFile)
        tmpHdfPath = pathlib.Path(self.path) / tmpHdfFile

        hdfFile = os.path.splitext(self.dbFile)[0] + ".h5"
        # hdfPath = os.path.join(self.path, hdfFile)
        hdfPath = pathlib.Path(self.path) / hdfFile
        logger.info(f"Rebuilding h5 to {hdfPath}")

        # can't pass sys.argv a 'PosixPath' from pathlib.Path, needs to be a string
        tmpHdfPath = str(tmpHdfPath)
        hdfPath = str(hdfPath)

        # when calling ptrepack, we need trailing ':' on each src/dst path
        # without this Windows fails to find the file
        _tmpHdfPath = tmpHdfPath + ":"
        _hdfPath = hdfPath + ":"

        # The first item is normally the command line command name (not used)
        sys.argv = ["", "--overwrite", "--chunkshape=auto", _tmpHdfPath, _hdfPath]

        logger.info("running tables.scripts.ptrepack.main()")
        logger.info(f"sys.argv: {sys.argv}")
        try:
            tables.scripts.ptrepack.main()  # noqa

            # delete the temporary (large file)
            logger.info(f"Deleting tmp file: {tmpHdfPath}")
            os.remove(tmpHdfPath)

            self.signalApp(
                f"Saved compressed folder analysis with tables.scripts.ptrepack.main()"
            )

        except FileNotFoundError as e:
            logger.error("tables.scripts.ptrepack.main() failed ... file was not saved")
            logger.error(e)
            self.signalApp(f"ERROR in tables.scripts.ptrepack.main(): {e}")

    def _not_used_save(self):
        """Save all analysis as csv."""
        logger.info("")
        for ba in self:
            if ba is not None:
                # save detection and analysis to json
                # ba.saveAnalysis(forceSave=True)

                # save analysis to csv
                ba.saveAnalysis_tocsv()

    def old_getTmpHdfFile(self):
        """Get temporary h5 file to write to.

        We will always then compress with _rebuildHdf.
        """
        logger.info("")

        tmpHdfFile = os.path.splitext(self.dbFile)[0] + "_tmp.h5"
        tmpHdfPath = pathlib.Path(self.path) / tmpHdfFile

        # the compressed version from the last save
        hdfFile = os.path.splitext(self.dbFile)[0] + ".h5"
        hdfFilePath = pathlib.Path(self.path) / hdfFile

        # hdfMode = 'w'
        if os.path.isfile(hdfFilePath):
            logger.info(f"    copying existing hdf file to tmp ")
            logger.info(f"    hdfFilePath {hdfFilePath}")
            logger.info(f"    tmpHdfPath {tmpHdfPath}")
            shutil.copyfile(hdfFilePath, tmpHdfPath)  # noqa
        else:
            pass
            # compressed file does not exist, just use tmp path
            # print('   does not exist:', hdfFilePath)

        return tmpHdfPath

    def getPathFromRelPath(self, relPath):
        """Get full path to file (usually an abf file."""
        if relPath.startswith("/"):
            relPath = relPath[1:]

        fullFilePath = os.path.join(self.path, relPath)

        """
        print('xxx', self.path)
        print('xxx', relPath)
        print('xxx', fullFilePath)
        """

        return fullFilePath

    def saveHdf(self):
        """
        Save file table and any number of loaded and analyzed bAnalysis.

        Set file table 'uuid' column when we actually save a bAnalysis

        Important: Order matters
                (1) Save bAnalysis first, it updates uuid in file table.
                (2) Save file table with updated uuid
        """
        start = time.time()

        df = self.getDataFrame()

        # the compressed version from the last save
        hdfFile = os.path.splitext(self.dbFile)[0] + ".h5"
        hdfFilePath = pathlib.Path(self.path) / hdfFile

        logger.info(f"Saving db (will be compressed) {hdfFilePath}")

        # save each bAnalysis
        for row in range(len(df)):
            ba = df.at[row, "_ba"]
            if ba is not None:
                didSave = ba._saveHdf_pytables(hdfFilePath)
                if didSave:
                    # we are now saved into h5 file, remember uuid to load
                    # print('xxx SETTING dir uuid')
                    df.at[row, "uuid"] = ba.uuid

        # rebuild (L, A, S) columns
        self._updateLoadedAnalyzed()

        #
        # save file database
        logger.info(f"    saving file db with {len(df)} rows")
        print(df)

        dbKey = os.path.splitext(self.dbFile)[0]
        df = df.drop("_ba", axis=1)  # don't ever save _ba, use it for runtime

        # hdfStore[dbKey] = df  # save it
        df.to_hdf(hdfFilePath, dbKey)

        #
        self._isDirty = False  # if true, prompt to save on quit

        # rebuild the file to remove old changes and reduce size
        # self._rebuildHdf()
        sanpy.h5Util._repackHdf(hdfFilePath)

        # list the keys in the file
        sanpy.h5Util.listKeys(hdfFilePath)

        stop = time.time()
        logger.info(f"Saving took {round(stop-start,2)} seconds")

    def loadHdf(self, path=None):
        """
        Load the database key from an h5 file.

        We do not load analy anlysis until user clicks on row, see loadOneAnalysis()
        """
        if path is None:
            path = self.path
        self.path = path

        df = None
        hdfFile = os.path.splitext(self.dbFile)[0] + ".h5"
        hdfPath = pathlib.Path(self.path) / hdfFile
        if not hdfPath.is_file():
            return

        logger.info(f"Loading existing folder h5 file {hdfPath}")
        sanpy.h5Util.listKeys(hdfPath)

        _start = time.time()
        dbKey = os.path.splitext(self.dbFile)[0]

        try:
            df = pd.read_hdf(hdfPath, dbKey)
        except KeyError as e:
            # file is corrupt !!!
            logger.error(f'    Load h5 failed, did not find dbKey:"{dbKey}" {e}')

        if df is not None:
            # _ba is for runtime, assign after loading from either (abf or h5)
            df["_ba"] = None

            logger.info("    loaded db df")
            logger.info(f"{df[['File', 'uuid']]}")

            # do not load anything until user clicks rows, see loadOneAnalysis()

            _stop = time.time()
            logger.info(f"Loading took {round(_stop-_start,2)} seconds")
        #
        return df

    def loadOneAnalysis(self, path, uuid=None, allowAutoLoad=True, verbose=True):
        """Load one bAnalysis either from original file path or uuid of h5 file.

        If from h5, we still need to reload sweeps !!!
        They are binary and fast, saving to h5 (in this case) is slow.
        """
        if verbose:
            logger.info(f'path:"{path}" uuid:"{uuid}" allowAutoLoad:"{allowAutoLoad}"')

        hdfPath = self._getHdfFile()

        # grab the fileLoaderDict from our app
        # if it is None then bAnalysis will load this (from disk)
        if self.myApp is not None:
            _fileLoaderDict = self.myApp.getFileLoaderDict()
        else:
            _fileLoaderDict = None

        ba = None
        if uuid is not None and uuid:
            # load from h5
            if verbose:
                logger.info(f"    Retreiving uuid from hdf file {uuid}")

            # load from abf
            ba = sanpy.bAnalysis(path, fileLoaderDict=_fileLoaderDict, verbose=verbose)

            # load analysis from h5 file, will fail if uuid is not in file
            ba._loadHdf_pytables(hdfPath, uuid)

        if allowAutoLoad and ba is None:
            # load from path
            ba = sanpy.bAnalysis(path, fileLoaderDict=_fileLoaderDict, verbose=verbose)
            if verbose:
                logger.info(f"    Loaded ba from path {path} and now ba:{ba}")
        #
        return ba

    def _getHdfFile(self):
        hdfFile = os.path.splitext(self.dbFile)[0] + ".h5"
        hdfPath = os.path.join(self.path, hdfFile)
        return hdfPath

    def _deleteFromHdf(self, uuid):
        """Delete uuid from h5 file.

        Each bAnalysis detection get a unique uuid.
        """
        if uuid is None or not uuid:
            return
        logger.info(f"deleting from h5 file uuid:{uuid}")

        _hdfFile = os.path.splitext(self.dbFile)[0] + ".h5"
        hdfPath = pathlib.Path(self.path) / _hdfFile

        # tmpHdfPath = self._getTmpHdfFile()

        removed = False
        with pd.HDFStore(hdfPath) as hdfStore:
            try:
                hdfStore.remove(uuid)
                removed = True
            except KeyError:
                logger.error(f"Did not find uuid {uuid} in h5 file {hdfPath}")

        #
        if removed:
            # will rebuild on next save
            # self._rebuildHdf()
            self._updateLoadedAnalyzed()
            self._isDirty = True  # if true, prompt to save on quit

    def loadFolder(self, path=None, loadData=False):
        """
        Parse a folder and load all (abf, csv, ...). Only called if no h5 file.

        TODO: get rid of loading database from .csv (it is replaced by .h5 file)
        TODO: extend the logic to load from cloud (after we were instantiated)
        """
        logger.info("Loading folder from scratch (no hdf file)")

        start = time.time()
        if path is None:
            path = self.path
        self.path = path

        loadedDatabase = False

        # load an existing folder db or create a new one
        # abb 20220612 turned off loading from self.dbFile .csv file
        dbPath = os.path.join(path, self.dbFile)
        if 0 and os.path.isfile(dbPath):
            # load from .csv
            logger.info(f"Loading existing folder db: {dbPath}")
            df = pd.read_csv(dbPath, header=0, index_col=False)
            # df["Idx"] = pd.to_numeric(df["Idx"])
            df = self._setColumnType(df)
            loadedDatabase = True
            # logger.info(f'  shape is {df.shape}')
        else:
            # logger.info(f'No existing db file, making {dbPath}')
            logger.info(f"No existing db file, making default dataframe")
            df = pd.DataFrame(columns=self.sanpyColumns.keys())
            df = self._setColumnType(df)

        if loadedDatabase:
            # check columns with sanpyColumns
            loadedColumns = df.columns
            for col in loadedColumns:
                if not col in self.sanpyColumns.keys():
                    logger.error(
                        f'error: bAnalysisDir did not find loaded col: "{col}" in sanpyColumns.keys()'
                    )
            for col in self.sanpyColumns.keys():
                if not col in loadedColumns:
                    logger.error(
                        f'error: bAnalysisDir did not find sanpyColumns.keys() col: "{col}" in loadedColumns'
                    )

        if loadedDatabase:
            # seach existing db for missing abf files
            pass
        else:
            # get list of all abf/csv/tif files
            fileList = self.getFileList(path)
            _numFilesToLoad = len(fileList)
            start = time.time()
            # build new db dataframe
            listOfDict = []
            for rowIdx, fullFilePath in enumerate(fileList):
                self.signalApp(
                    f'Loading file {rowIdx+1} of {_numFilesToLoad} "{fullFilePath}"'
                )

                # rowDict is what we are showing in the file table
                # abb debug vue, set loadData=True
                # loads bAnalysis
                ba, rowDict = self.getFileRow(fullFilePath, loadData=loadData)

                if rowDict is None:
                    logger.warning(f'error loading file {fullFilePath}')
                    continue
                
                # TODO: calculating time, remove this
                # This is 2x faster than loading from pandas gzip ???
                # dDict = sanpy.bAnalysis.getDefaultDetection()
                # dDict['dvdtThreshold'] = 2
                # ba.spikeDetect(dDict)

                # as we parse the folder, don't load ALL files (will run out of memory)
                if loadData:
                    rowDict["_ba"] = ba
                else:
                    rowDict["_ba"] = None  # ba

                # do not assign uuid until bAnalysis is saved in h5 file
                # rowDict['uuid'] = ''

                # 20230313 moved into getFileRow()
                # relPath = fullFilePath.replace(path, '')
                # if relPath.startswith('/'):
                #     # so we can use os.path.join()
                #     relPath = relPath[1:]
                # rowDict['relPath'] = relPath

                # logger.info(f'    row:{rowIdx} relPath:{relPath} fullFilePath:{fullFilePath}')

                listOfDict.append(rowDict)

            stop = time.time()
            logger.info(f"Load took {round(stop-start,3)} seconds.")

            #
            df = pd.DataFrame(listOfDict)
            # print('=== built new db df:')
            # print(df)
            df = self._setColumnType(df)
        #

        # expand each to into self.fileList
        # df['_ba'] = None

        stop = time.time()
        logger.info(f"Load took {round(stop-start,2)} seconds.")
        return df

    def _checkColumns(self):
        """Check columns in loaded vs sanpyColumns (and vica versa).
        """
        if self._df is None:
            return
        loadedColumns = self._df.columns
        for col in loadedColumns:
            if not col in self.sanpyColumns.keys():
                # loaded has unexpected column, leave it
                logger.info(
                    f'did not find loaded col: "{col}" in sanpyColumns.keys() ... ignore it'
                )
        for col in self.sanpyColumns.keys():
            if not col in loadedColumns:
                # loaded is missing expected, add it
                logger.info(
                    f'did not find sanpyColumns.keys() col: "{col}" in loadedColumns ... adding col'
                )
                self._df[col] = ""

    def _updateLoadedAnalyzed(self, theRowIdx=None):
        """Refresh Loaded (L) and Analyzed (A) columns.

        Arguments:
            theRowIdx (int): Update just one row

        TODO: For kymograph, add rows (left, top, right, bottom) and update
        """
        if self._df is None:
            return
        for rowIdx in range(len(self._df)):
            if theRowIdx is not None and theRowIdx != rowIdx:
                continue

            ba = self._df.loc[rowIdx, "_ba"]  # Can be None

            # uuid = self._df.at[rowIdx, 'uuid']
            #
            # loaded
            if self.isLoaded(rowIdx):
                theChar = "\u2022"  # FILLED BULLET
            # elif uuid:
            #    #theChar = '\u25CB'  # open circle
            #    theChar = '\u25e6'  # white bullet
            else:
                theChar = ""
            # self._df.iloc[rowIdx, loadedCol] = theChar
            self._df.loc[rowIdx, "L"] = theChar
            #
            # analyzed
            if self.isAnalyzed(rowIdx):
                theChar = "\u2022"  # FILLED BULLET
                self._df.loc[rowIdx, "N"] = ba.numSpikes
                _numErrors = ba.numErrors
                if _numErrors is None:
                    _numErrors = ''
                logger.warning(f'setting E to _numErrors {_numErrors}')
                self._df.loc[rowIdx, "E"] = _numErrors
            # elif uuid:
            #    #theChar = '\u25CB'
            #    theChar = '\u25e6'  # white bullet
            else:
                theChar = ""
                self._df.loc[rowIdx, "A"] = ""
            # self._df.iloc[rowIdx, analyzedCol] = theChar
            self._df.loc[rowIdx, "A"] = theChar
            #
            # saved
            if self.isSaved(rowIdx):
                theChar = "\u2022"  # FILLED BULLET
            else:
                theChar = ""
            # self._df.iloc[rowIdx, savedCol] = theChar
            self._df.loc[rowIdx, "S"] = theChar
            #
            # start(s) and stop(s) from ba detectionDict
            if self.isAnalyzed(rowIdx):
                # set table to values we just detected with
                startSec = ba.getDetectionDict()["startSeconds"]
                stopSec = ba.getDetectionDict()["stopSeconds"]
                self._df.loc[rowIdx, "Start(s)"] = startSec
                self._df.loc[rowIdx, "Stop(s)"] = stopSec

                dvdtThreshold = ba.getDetectionDict()["dvdtThreshold"]
                mvThreshold = ba.getDetectionDict()["mvThreshold"]
                self._df.loc[rowIdx, "dvdtThreshold"] = dvdtThreshold
                self._df.loc[rowIdx, "mvThreshold"] = mvThreshold

                #
                # TODO: remove start of ba._path that corresponds to our current folder path
                # will allow our save db to be modular

                # relPth should usually be filled in ???
                """
                relPath = self.getPathFromRelPath(ba._path)
                self._df.loc[rowIdx, 'relPath'] = relPath
                """

                # logger.info('maybe put back in')
                # print(f'    self._df.loc[rowIdx, "relPath"] is "{self._df.loc[rowIdx, "relPath"]}"')

            # kymograph interface
            # if ba is not None and ba.isKymograph():
            # if ba is not None and isinstance(ba, sanpy.fileloaders.fileLoader_tif):
            if ba is not None and ba.fileLoader.isKymograph():
                kRect = ba.fileLoader.getKymographRect()

                # print(kRect)
                # sys.exit(1)

                if kRect is None:
                    logger.error(f"Got None kymograph rect")
                else:
                    self._df.loc[rowIdx, "kLeft"] = kRect[0]
                    self._df.loc[rowIdx, "kTop"] = kRect[1]
                    self._df.loc[rowIdx, "kRight"] = kRect[2]
                    self._df.loc[rowIdx, "kBottom"] = kRect[3]
                #
                # TODO: remove start of ba._path that corresponds to our current folder path
                # will allow our save db to be modular
                # self._df.loc[rowIdx, 'path'] = ba._path

    """
    def setCellValue(self, rowIdx, colStr, value):
        self._df.loc[rowIdx, colStr] = value
    """

    def isLoaded(self, rowIdx):
        isLoaded = self._df.loc[rowIdx, "_ba"] is not None
        return isLoaded

    def isAnalyzed(self, rowIdx):
        isAnalyzed = False
        ba = self._df.loc[rowIdx, "_ba"]
        # print('isAnalyzed()', rowIdx, ba)
        # if ba is not None:
        # print('qqq', rowIdx, ba, type(ba))
        # sanpy.bAnalysis_.bAnalysis
        # if isinstance(ba, sanpy.bAnalysis):
        if ba is not None:
            isAnalyzed = ba.isAnalyzed()
        return isAnalyzed

    def analysisIsDirty(self, rowIdx):
        """Analysis is dirty when there has been detection but not saved to h5."""
        isDirty = False
        ba = self._df.loc[rowIdx, "_ba"]
        if isinstance(ba, sanpy.bAnalysis):
            isDirty = ba.isDirty()
        return isDirty

    def hasDirty(self):
        """Return true if any bAnalysis in list has been analyzed but not saved (e.g. is dirty)"""
        haveDirty = False
        numRows = len(self._df)
        for rowIdx in range(numRows):
            if self.analysisIsDirty(rowIdx):
                haveDirty = True

        return haveDirty

    def isSaved(self, rowIdx):
        uuid = self._df.at[rowIdx, "uuid"]
        return len(uuid) > 0

    def getAnalysis(self, rowIdx, allowAutoLoad=True, verbose=True) -> sanpy.bAnalysis:
        """Get bAnalysis object, will load if necc.

        Args:
            rowIdx (int): Row index from table, corresponds to row in self._df
            allowAutoLoad (bool)
        Return:
            bAnalysis
        """
        file = self._df.loc[rowIdx, "File"]
        ba = self._df.loc[rowIdx, "_ba"]
        uuid = self._df.loc[
            rowIdx, "uuid"
        ]  # if we have a uuid bAnalysis is saved in h5f
        # filePath = os.path.join(self.path, file)
        # logger.info(f'Found _ba in file db with ba:"{ba}" {type(ba)}')
        # logger.info(f'rowIdx: {rowIdx} ba:{ba}')

        if ba is None or ba == "":
            # logger.info('did not find _ba ... loading from abf file ...')
            # working on kymograph
            #                 relPath = self.getPathFromRelPath(ba._path)
            relPath = self._df.loc[rowIdx, "relPath"]
            filePath = self.getPathFromRelPath(relPath)

            ba = self.loadOneAnalysis(
                filePath, uuid, allowAutoLoad=allowAutoLoad, verbose=verbose
            )
            # load
            """
            logger.info(f'Loading bAnalysis from row {rowIdx} "{filePath}"')
            ba = sanpy.bAnalysis(filePath)
            """
            if ba is None:
                logger.warning(
                    f'Did not load row {rowIdx} path: "{filePath}". Analysis was probably not saved'
                )
            else:
                self._df.at[rowIdx, "_ba"] = ba
                # does not get a uuid until save into h5
                if uuid:
                    # there was an original uuid (in table), means we are saved into h5
                    self._df.at[rowIdx, "uuid"] = uuid
                    if uuid != ba.uuid:
                        logger.error(
                            "Loaded uuid does not match existing in file table"
                        )
                        logger.error(f"  Loaded {ba.uuid}")
                        logger.error(f"  Existing {uuid}")

                # kymograph, set ba rect from table
                # if ba.isKymograph():
                # if ba is not None and isinstance(ba, sanpy.fileloaders.fileLoader_tif):
                if ba is not None and ba.fileLoader.isKymograph():
                    left = self._df.loc[rowIdx, "kLeft"]
                    top = self._df.loc[rowIdx, "kTop"]
                    right = self._df.loc[rowIdx, "kRight"]
                    bottom = self._df.loc[rowIdx, "kBottom"]

                    # on first load, these will be empty
                    # grab rect from ba (in _updateLoadedAnalyzed())
                    if left == "" or top == "" or right == "" or bottom == "":
                        pass
                    else:
                        theRect = [left, top, right, bottom]
                        logger.info(f"  theRect:{theRect}")
                        ba.fileLoader._updateTifRoi(theRect)

                #
                # update stats of table load/analyzed columns
                self._updateLoadedAnalyzed()

        return ba

    def _setColumnType(self, df):
        """Needs to be called every time a df is created.
        Ensures proper type of columns following sanpyColumns[key]['type']
        """
        # print('columns are:', df.columns)
        for col in df.columns:
            # when loading from csv, 'col' may not be in sanpyColumns
            if not col in self.sanpyColumns:
                logger.warning(f'Column "{col}" is not in sanpyColumns -->> ignoring')
                continue
            colType = self.sanpyColumns[col]["type"]
            # print(f'  _setColumnType() for "{col}" is type "{colType}"')
            # print(f'    df[col]:', 'len:', len(df[col]))
            # print(df[col])
            if colType == str:
                df[col] = df[col].replace(np.nan, "", regex=True)
                df[col] = df[col].astype(str)
            elif colType == int:
                pass
                # print('!!! df[col]:', df[col])
                # df[col] = df[col].astype(int)
            elif colType == float:
                # error if ''
                df[col] = df[col].astype(float)
            elif colType == bool:
                df[col] = df[col].astype(bool)
            else:
                logger.warning(f'Did not parse col "{col}" with type "{colType}"')
        #
        return df

    def getFileRow(self, path, loadData=False):
        """Get dict representing one file (row in table). Loads bAnalysis to get headers.

        On load error of proper file type (abf, csv), ba.loadError==True

        Args:
            path (Str): Full path to file.
            #rowIdx (int): Optional row index to assign in column 'Idx'

        Return:
            (tuple): tuple containing:

            - ba (bAnalysis): [sanpy.bAnalysis](/api/bAnalysis).
            - rowDict (dict): On success, otherwise None.
                    fails when path does not lead to valid bAnalysis file.
        """
        if not os.path.isfile(path):
            logger.warning(f'Did not find file "{path}"')
            return None, None
        fileType = os.path.splitext(path)[1]
        if fileType not in self.theseFileTypes:
            logger.warning(f'Did not load file type "{fileType}"')
            return None, None

        # grab the fileLoaderDict from our app
        # if it is None then bAnalysis will load this (from disk)
        if self.myApp is not None:
            _fileLoaderDict = self.myApp.getFileLoaderDict()
        else:
            _fileLoaderDict = None

        # load bAnalysis
        # logger.info(f'Loading bAnalysis "{path}"')
        # loadData is false, load header
        ba = sanpy.bAnalysis(path, loadData=loadData, fileLoaderDict=_fileLoaderDict)

        if ba.loadError:
            logger.error(f'Error loading bAnalysis file "{path}"')
            # return None, None

        # not sufficient to default everything to empty str ''
        # sanpyColumns can only have type in ('float', 'str')
        rowDict = dict.fromkeys(self.sanpyColumns.keys(), "")
        for k in rowDict.keys():
            if self.sanpyColumns[k]["type"] == str:
                rowDict[k] = ""
            elif self.sanpyColumns[k]["type"] == float:
                rowDict[k] = np.nan

        # if rowIdx is not None:
        #    rowDict['Idx'] = rowIdx

        """
        if ba.loadError:
            rowDict['I'] = 0
        else:
            rowDict['I'] = 2 # need 2 because checkbox value is in (0,2)
        """

        if ba.loadError:
            return None, None
        
        rowDict["File"] = ba.fileLoader.filename  # os.path.split(ba.path)[1]
        rowDict["Dur(s)"] = ba.fileLoader.recordingDur

        rowDict["Channels"] = ba.fileLoader.numChannels  # Theanne

        rowDict["Sweeps"] = ba.fileLoader.numSweeps

        # TODO: here, we do not get an epoch table until the file is loaded !!!
        rowDict["Epochs"] = ba.fileLoader.numEpochs  # Theanne, data has to be loaded

        rowDict["kHz"] = ba.fileLoader.recordingFrequency
        rowDict["Mode"] = ba.fileLoader.recordingMode.value

        # rowDict['dvdtThreshold'] = 20
        # rowDict['mvThreshold'] = -20
        if ba.isAnalyzed():
            dDict = ba.getDetectionDict()
            # rowDict['I'] = dDict.getValue('include')
            rowDict["dvdtThreshold"] = dDict.getValue("dvdtThreshold")
            rowDict["mvThreshold"] = dDict.getValue("mvThreshold")
            rowDict["Start(s)"] = dDict.getValue("startSeconds")
            rowDict["Stop(s)"] = dDict.getValue("stopSeconds")

        # remove the path to the folder we have loaded
        relPath = path.replace(self.path, "")
        
        # logger.info(f'xxx self.path: "{self.path}"')
        # logger.info(f'xxx path: "{path}"')
        # logger.info(f'xxx relPath: "{relPath}"')
        
        if relPath.startswith("/"):
            # so we can use os.path.join()
            relPath = relPath[1:]
        # added 20230505 working with johnson in 1313 to fix windows bug ???
        if relPath.startswith("\\"):
            # so we can use os.path.join()
            relPath = relPath[1:]

        rowDict["relPath"] = relPath

        #logger.info(f'2) xxx relPath: "{relPath}"')

        return ba, rowDict

    def getFileList(self, path: str = None) -> List[str]:
        """Get file paths from path.

        Uses self.theseFileTypes
        """
        if path is None:
            path = self.path

        logger.warning("Remember: MODIFIED TO LOAD TIF FILES IN SUBFOLDERS")
        count = 1
        tmpFileList = []
        folderDepth = self.folderDepth  # if none then all depths
        excludeFolders = ["analysis", "hide"]
        for root, subdirs, files in os.walk(path):
            subdirs[:] = [d for d in subdirs if d not in excludeFolders]

            # print('folderDepth:', folderDepth)
            # print('  root:', root, 'subdirs:', subdirs, 'files:', files)

            # strip out folders that start with __
            # _parentFolder = os.path.split(root)[1]
            # print('root:', root)
            # print('  parentFolder:', _parentFolder)
            # if _parentFolder.startswith('__'):
            if "__" in root:
                logger.info(f"SKIPPING based on path root:{root}")
                continue

            if os.path.split(root)[1] == "analysis":
                # don't load from analysis/ folder, we save analysis there
                continue

            # if os.path.split(root)[1] == 'hide':
            #     # special case/convention, don't load from 'hide' folders
            #     continue

            for file in files:
                # TODO (cudmore) parse all our fileLoader(s) for a list
                _, _ext = os.path.splitext(file)
                if _ext in self.theseFileTypes:
                    oneFile = os.path.join(root, file)
                    tmpFileList.append(oneFile)

            count += 1
            if folderDepth is not None and count > folderDepth:
                break

        fileList = []
        for file in sorted(tmpFileList):
            if file.startswith("."):
                continue
            # ignore our database file
            if file == self.dbFile:
                continue

            # tmpExt is like .abf, .csv, etc
            tmpFileName, tmpExt = os.path.splitext(file)
            if tmpExt in self.theseFileTypes:
                # if getFullPath:
                #     #file = os.path.join(path, file)
                #     file = pathlib.Path(path) / file
                #     file = str(file)  # return List[str] NOT List[PosixPath]
                fileList.append(file)
        #
        logger.info(f"found {len(fileList)} files ...")
        return fileList

    def getRowDict(self, rowIdx):
        """
        Return a dict with selected row as dict (includes detection parameters).

        Important to return a copy as our '_ba' is a pointer to bAnalysis.

        Returns:
            theRet (dict): Be sure to make a deep copy of ['_ba'] if neccessary.
        """
        theRet = {}
        # use columns in main sanpyColumns, not in df
        # for colStr in self.columns:
        for colStr in self._df.columns:
            # theRet[colStr] = self._df.loc[rowIdx, colStr]
            theRet[colStr] = self._df.loc[rowIdx, colStr]
        # theRet['_ba'] = theRet['_ba'].copy()
        return theRet

    def appendRow(self, rowDict=None, ba=None):
        """Append an empty row."""

        # logger.info('')
        # print('    rowDict:', rowDict)
        # print('    ba:', ba)

        rowSeries = pd.Series()
        if rowDict is not None:
            rowSeries = pd.Series(rowDict)
            # self._data.iloc[row] = rowSeries
            # self._data = self._data.reset_index(drop=True)

        newRowIdx = len(self._df)
        df = self._df
        logger.warning(f"need to replace append with concat")
        df = df.append(rowSeries, ignore_index=True)
        # df = pd.concat([df,rowSeries], ignore_index=True, axis=1)
        df = df.reset_index(drop=True)

        if ba is not None:
            df.loc[newRowIdx, "_ba"] = ba

        #
        self._df = df

    def unloadRow(self, rowIdx):
        self._df.loc[rowIdx, "_ba"] = None
        self._updateLoadedAnalyzed()

    def removeRowFromDatabase(self, rowIdx):
        # delete from h5 file
        uuid = self._df.at[rowIdx, "uuid"]
        self._deleteFromHdf(uuid)

        # clear uuid
        self._df.at[rowIdx, "uuid"] = ""

        self._updateLoadedAnalyzed()

    def deleteRow(self, rowIdx):
        df = self._df

        # delete from h5 file
        uuid = df.at[rowIdx, "uuid"]
        self._deleteFromHdf(uuid)

        # delete from df/model
        df = df.drop([rowIdx])
        df = df.reset_index(drop=True)
        self._df = df

        self._updateLoadedAnalyzed()

    def old_duplicateRow(self, rowIdx):
        """Depreciated, Was used to have different ocnditions within a recording,
        this is now handled by condiiton column.
        """
        # duplicate rowIdx
        newIdx = rowIdx + 0.5

        rowDict = self.getRowDict(rowIdx)

        # CRITICAL: Need to make a deep copy of the _ba pointer to bAnalysis object
        logger.info(f"copying {type(rowDict['_ba'])} {rowDict['_ba']}")
        baNew = copy.deepcopy(rowDict["_ba"])

        # copy of bAnalysis needs a new uuid
        new_uuid = (
            sanpy._util.getNewUuid()
        )  # 't' + str(uuid.uuid4())   #.replace('-', '_')
        logger.info(f"assigning new uuid {new_uuid} to {baNew}")

        if baNew.uuid == new_uuid:
            logger.error("!!!!!!!!!!!!!!!!!!!!!!!!!CRITICAL, new uuid is same as old")

        baNew.uuid = new_uuid

        rowDict["_ba"] = baNew
        rowDict["uuid"] = baNew.uuid  # new row can never have same uuid as old

        dfRow = pd.DataFrame(rowDict, index=[newIdx])

        df = self._df
        df = df.append(dfRow, ignore_index=True)
        df = df.sort_values(by=["File"], axis="index", ascending=True, inplace=False)
        df = df.reset_index(drop=True)
        self._df = df

        self._updateLoadedAnalyzed()

    def syncDfWithPath(self):
        """Sync path with existing df. Used to detect new/removed files."""

        pathFileList = self.getFileList()  # always full path
        dfFileList = self._df["File"].tolist()

        logger.info("")
        # print('    === pathFileList (on drive):')
        # print('    ', pathFileList)
        # print('    === dfFileList (in table):')
        # print('    ', dfFileList)

        addedToDf = False

        # look for files in path not in df
        for pathFile in pathFileList:
            fileName = os.path.split(pathFile)[1]
            if fileName not in dfFileList:
                logger.info(f'Found file in path "{fileName}" not in df')
                # load bAnalysis and get df column values
                addedToDf = True
                # fullPathFile = os.path.join(self.path, pathFile)
                ba, rowDict = self.getFileRow(pathFile)  # loads bAnalysis
                if rowDict is not None:
                    # listOfDict.append(rowDict)

                    # TODO: get this into getFileROw()
                    logger.warning("20220718, not sure we need this ???")
                    # rowDict['relPath'] = pathFile
                    rowDict["_ba"] = None

                    self.appendRow(rowDict=rowDict, ba=None)

        # look for files in df not in path
        # for dfFile in dfFileList:
        #     if not dfFile in pathFileList:
        #         logger.info(f'Found file in df "{dfFile}" not in path')

        if addedToDf:
            df = self._df
            df = df.sort_values(
                by=["File"], axis="index", ascending=True, inplace=False
            )
            df = df.reset_index(drop=True)
            self._df = df

        self._updateLoadedAnalyzed()

    def pool_build(self):
        """Build one df with all analysis. Use this in plot tool plugin."""
        logger.info("")
        masterDf = None
        for row in range(self.numFiles):
            if not self.isAnalyzed(row):
                logger.info(f"  row:{row} not analyzed")
                continue
            ba = self.getAnalysis(row)
            oneDf = ba.asDataFrame()
            if oneDf is not None:
                self.signalApp(f'  adding "{ba.fileLoader.filename}"')
                oneDf["File Number"] = int(row)
                if masterDf is None:
                    masterDf = oneDf
                else:
                    masterDf = pd.concat([masterDf, oneDf])
        #
        if masterDf is None:
            logger.error("Did not find any analysis.")
        else:
            logger.info(f"final num spikes {len(masterDf)}")
        # print(masterDf.head())
        self._poolDf = masterDf

        return self._poolDf

    def signalApp(self, str):
        """Update status bar of SanPy app.

        TODO make this a signal and connect app to it.
            Will not be able to do this, we need to run outside Qt
        """
        if self.myApp is not None:
            self.myApp.slot_updateStatus(str)
        else:
            logger.info(str)

    def api_getFileHeaders(self):
        headerList = []
        df = self.getDataFrame()
        for row in range(len(df)):
            # ba = self.getAnalysis(row)  # do not call this, it will load
            ba = df.at[row, "_ba"]
            if ba is not None:
                headerDict = ba.api_getHeader()
                headerList.append(headerDict)
        #
        return headerList


def _printDict(d):
    for k, v in d.items():
        print("  ", k, ":", v)


def test3():
    path = "/home/cudmore/Sites/SanPy/data"
    bad = analysisDir(path)

    # file = '19221014.abf'
    rowIdx = 3
    ba = bad.getAnalysis(rowIdx)
    ba = bad.getAnalysis(rowIdx)

    print(bad.getDataFrame())

    print("bad.shape", bad.shape)
    print("bad.columns:", bad.columns)

    print("bad.iloc[rowIdx,5]:", bad.iloc[rowIdx, 5])
    print("setting to xxxyyyzzz")

    # setter
    # bad.iloc[2,5] = 'xxxyyyzzz'

    print("bad.iloc[2,5]:", bad.iloc[rowIdx, 5])
    print('bad.loc[2,"File"]:', bad.loc[rowIdx, "File"])

    print("bad.iloc[2]")
    print(bad.iloc[rowIdx])
    # bad.iloc[rowIdx] = ''
    print("bad.loc[2]")
    print(bad.loc[rowIdx])

    # bad.saveDatabase()


def test_hd5_2():
    folderPath = "/home/cudmore/Sites/SanPy/data"
    if 1:
        # save analysisDir hdf
        bad = analysisDir(folderPath)
        print("bad._df:")
        print(bad._df)
        # bad.saveDatabase()  # save .csv
        bad.saveHdf()  # save ALL bAnalysis in .h5

    if 0:
        # load h5 and reconstruct a bAnalysis object
        start = time.time()
        hdfPath = "/home/cudmore/Sites/SanPy/data/sanpy_recording_db.h5"
        with pd.HDFStore(hdfPath, "r") as hdfStore:
            for key in hdfStore.keys():
                dfTmp = hdfStore[key]
                # path = dfTmp.iloc[0]['path']

                print("===", key)
                # if not key.startswith('/r6'):
                #    continue
                # for col in dfTmp.columns:
                #    print('  ', col, dfTmp[col])
                # print(key, type(dfTmp), dfTmp.shape)
                # print('dfTmp:')
                # print(dfTmp)

                # print(dfTmp.iloc[0]['_sweepX'])

                # load bAnalysis from a pandas DataFrame
                ba = sanpy.bAnalysis(fromDf=dfTmp)

                print(ba)

                ba.spikeDetect()  # this should reproduce exactly what was save ... It WORKS !!!

        stop = time.time()
        logger.info(f"h5 load took {round(stop-start,3)} seconds")


def test_hd5():
    import time

    start = time.time()
    hdfStore = pd.HDFStore("store.gzip")
    if 1:
        path = "/home/cudmore/Sites/SanPy/data"
        bad = analysisDir(path)

        # load all bAnalysis
        for idx in range(len(bad)):
            bad.getAnalysis(idx)

        hdfStore["df"] = bad.getDataFrame()  # save it
        bad.saveDatabase()

    if 0:
        df = hdfStore["df"]  # load it
        print("loaded df:")
        print(df)
    stop = time.time()
    print(f"took {stop-start}")


def test_pool():
    path = "/home/cudmore/Sites/SanPy/data"
    ad = analysisDir(path)
    print("loaded df:")
    print(ad._df)

    ad.pool_build()


def testCloud():
    cloudDict = {"owner": "cudmore", "repo_name": "SanPy", "path": "data"}
    bad = bAnalysisDirWeb(cloudDict)


if __name__ == "__main__":
    # test3()
    # test_hd5()
    test_hd5_2()
    # test_pool()
    # testCloud()

    # test_timing()
    # plotTiming()



import os
import glob
import math
import enum
import inspect
from typing import Union, Dict, List, Tuple, Optional
from abc import ABC, abstractmethod

import numpy as np
import scipy.signal

import sanpy.fileloaders

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


def getFileLoaders(verbose: bool = False) -> dict:
    """Load file loaders from both

    1) Module sanpy.fileloaders

    2) Folder <user>/Documents/SanPy/File Loaders

    Each file loader is a class derived from [fileLoader_base](../../api/fileloader/fileLoader_base.md)

    See: sanpy.interface.bPlugins.loadPlugins()

    Returns
    -------
    dict
        A dictionary of file loaders.
    """
    retDict = {}

    ignoreModuleList = ["fileLoader_base", "recordingModes"]

    #
    # system file loaders from sanpy.fileloaders
    loadedList = []
    for moduleName, obj in inspect.getmembers(sanpy.fileloaders):
        if inspect.isclass(obj):
            if verbose:
                logger.info(f"moduleName:{moduleName}")
            if moduleName in ignoreModuleList:
                continue
            loadedList.append(moduleName)
            fullModuleName = "sanpy.fileloaders." + moduleName
            # filetype is a static str, e.g. the extension to load
            try:
                filetype = obj.loadFileType
            except AttributeError as e:
                logger.warning(f'Did not load "{moduleName}", no "filetype" attribute')
                continue
            oneLoaderDict = {
                "fileLoaderClass": moduleName,
                "type": "system",
                "module": fullModuleName,
                "path": "",
                "constructor": obj,
                #'filetype': filetype
            }
            if filetype in retDict.keys():
                logger.warning(
                    f'loader already added "{moduleName}" filetype:"{filetype}"'
                )
                logger.warning(f"  this loader will overwrite the previous loader.")
            retDict[filetype] = oneLoaderDict

    # logger.info(f'Loaded system file loaders:')
    # for k,v in retDict.keys():
    #     logger.info(f'    {k}:{v}')
    # # sort
    # retDict = dict(sorted(retDict.items()))

    #
    # user plugins from files in folder "<user>/SanPy/file loaders"
    fileLoaderFolder = sanpy._util._getUserFileLoaderFolder()
    loadedModuleList = []
    if os.path.isdir(fileLoaderFolder):
        files = glob.glob(os.path.join(fileLoaderFolder, "*.py"))
    else:
        # no user file loader folder ???
        files = []

    for file in files:
        if file.endswith("__init__.py"):
            continue

        moduleName = os.path.split(file)[1]
        moduleName = os.path.splitext(moduleName)[0]
        fullModuleName = "sanpy.fileloaders." + moduleName

        loadedModule = sanpy._util._module_from_file(fullModuleName, file)

        try:
            oneConstructor = getattr(loadedModule, moduleName)
        except AttributeError as e:
            logger.error(
                f'Did not load file loader, make sure file name and class name are the same:"{moduleName}"'
            )
        else:
            # filetype is a static str, e.g. the extension to load
            try:
                filetype = oneConstructor.loadFileType
            except AttributeError as e:
                logger.warning(f'Did not load "{moduleName}", no "filetype" attribute')
                continue
            oneLoaderDict = {
                "fileLoaderClass": moduleName,
                "type": "user",
                "module": fullModuleName,
                "path": file,
                "constructor": oneConstructor,
                #'filetype': filetype
            }
            if filetype in retDict.keys():
                logger.warning(
                    f'loader already added "{moduleName}" handleExtension:"{filetype}"'
                )
                logger.warning(f"  this loader will overwrite the previous loader.")
            retDict[filetype] = oneLoaderDict

    if verbose:
        logger.info(f"Loaded {len(retDict.keys())} file loaders:")
        for k, v in retDict.items():
            # logger.info(f'    {k}:{v}')
            logger.info(f"  {k}")
            for k2, v2 in v.items():
                logger.info(f"    {k2}: {v2}")

    # sort
    # retDict = dict(sorted(retDict.items()))

    return retDict


class recordingModes(enum.Enum):
    """Recording modes for I-Clamp, V-Clamp, and unknown."""

    iclamp = "I-Clamp"
    vclamp = "V-Clamp"
    kymograph = "Kymograph"
    unknown = "unknown"


class fileLoader_base(ABC):
    """Abstract base class to derive file loaders.

    For some working examples of derived classes, see
    [fileLoader_abf](../../api/fileloader/fileLoader_abf.md) and
    [fileLoader_csv](../../api/fileloader/fileLoader_csv.md)

    To create a file loader

    1) derive a class from fileLoader_base and define `loadFileType`

        class myFileLoader(fileLoader_base):
            loadFileType = 'the_file_extension_this_will_load'

    2) Define a `loadFile` function

        def loadFile(self):
            # load the data from self.filepath and create sweepX and sweepY
            # specify what was loaded
            self.setLoadedData(sweepX, sweepY)

    """

    loadFileType: str = ""
    # @property
    # @abstractmethod
    # def loadFileType(self) -> str:
    #     """Derived classes must return the file type to handle.
    #     For example, 'csv', or 'm', or 'dat'
    #     """
    #     pass

    @abstractmethod
    def loadFile(self):
        """Derived classes must load the data and call setLoadedData(sweepX, sweepY)."""
        pass

    def __init__(self, filepath: str, loadData: bool = True):
        """Base class to derive new file loaders.

        Parameters
        ----------
        filepath : str
            File path to load. Will use different derived classes based on extension
        loadData : bool
            If True then load raw data, otherwise just load the header.
        """

        super().__init__()

        self._loadError = False

        self._path = filepath

        self._filteredY : np.ndarray = None  # set in _getDerivative
        self._filteredDeriv : np.ndarray = None
        self._currentSweep: int = 0

        self._epochTableList: List[sanpy.fileloaders.epochTable] = None

        # load file from inherited class
        self.loadFile()

        # check our work
        self._checkLoadedData()

    def __str__(self):
        """Get a short string representing this file."""
        txt = f"file: {self.filename} sweeps: {self.numSweeps} dur (Sec):{self.recordingDur}"
        return txt

    def isKymograph(self) -> bool:
        return isinstance(self, sanpy.fileloaders.fileLoader_tif)

    @property
    def filepath(self) -> str:
        """Get the full file path."""
        return self._path

    @property
    def filename(self) -> str:
        """Get the filename."""
        if self.filepath is None:
            return None
        else:
            return os.path.split(self.filepath)[1]

    @property
    def numChannels(self) -> int:
        """Get the number of channels.

        If more than one channel, must be defined in derived class.
        """
        return 1

    @property
    def currentSweep(self) -> int:
        """Get the current sweep."""
        return self._currentSweep

    def setSweep(self, currentSweep: int):
        """Set the current sweep."""
        if currentSweep > self.numSweeps - 1:
            logger.error(f"max sweep is {self.numSweeps-1}, got {currentSweep}")
            return
        self._currentSweep = currentSweep

    @property
    def recordingMode(self):
        return self._recordingMode

    # feb 2023, uncommented
    @property
    def sweepLabelX(self):
        return self._sweepLabelX

    @property
    def sweepLabelY(self):
        return self._sweepLabelY

    @property
    def recordingDur(self):
        return self._sweepLengthSec

    @property
    def numSweeps(self):
        return len(self._sweepList)

    @property
    def sweepList(self):
        return self._sweepList

    @property
    def dataPointsPerMs(self):
        return self._dataPointsPerMs

    @property
    def acqDate(self):
        return self._acqDate

    @property
    def acqTime(self):
        return self._acqTime

    @property
    def sweepX(self):
        """Get the X-Values for a sweep.

        Notes
        -----
        All sweeps are assumed to have the same x-values (seconds).
        """
        # return self._sweepX[:, self.currentSweep]
        return self._sweepX[:, 0]

    @property
    def sweepY(self):
        """Get the Y values for the current sweep."""
        return self._sweepY[:, self.currentSweep]

    @property
    def sweepC(self):
        """Get the DAC command for the current sweep."""
        if self._sweepC is None:
            # return np.zeros_like(self._sweepX[:, self.currentSweep])
            return np.zeros_like(self._sweepX[:, 0])
        return self._sweepC[:, self.currentSweep]

    def get_xUnits(self):
        return self._sweepLabelX

    def get_yUnits(self):
        return self._sweepLabelY

    @property
    def filteredDeriv(self) -> Optional[np.ndarray]:
        """Get the filtered first derivative of sweepY."""
        if self._filteredDeriv is not None:
            return self._filteredDeriv[:, self.currentSweep]
        else:
            return None

    def _getDerivative(
        self,
        medianFilter: int = 0,
        SavitzkyGolay_pnts: int = 5,
        SavitzkyGolay_poly: int = 2,
    ):
        """Get filtered version of recording and derivative of recording (used for I-Clamp).

            By default we will use a SavitzkyGolay filter with 5 points and a 2nd order polynomial.

        Parameters
        ----------
        medianFilter : int
            Median filter box with. Must be odd, specify 0 for no median filter
        SavitzkyGolay_pnts : int
            Specify 0 for no filter.
        SavitzkyGolay_poly : int

        Notes
        -----
        Creates:
            self._filteredVm
            self._filteredDeriv
        """

        # logger.info(f'{self.filename} medianFilter:{medianFilter} SavitzkyGolay_pnts:{SavitzkyGolay_pnts} SavitzkyGolay_poly:{SavitzkyGolay_poly}')

        if not isinstance(medianFilter, int):
            logger.error(f"expecting int medianFilter, got: {medianFilter}")

        if medianFilter > 0:
            if not medianFilter % 2:
                medianFilter += 1
                logger.warning(
                    "Please use an odd value for the median filter, set medianFilter: {medianFilter}"
                )
            medianFilter = int(medianFilter)
            self._filteredY = scipy.signal.medfilt2d(self._sweepY, [medianFilter, 1])
        elif SavitzkyGolay_pnts > 0:
            self._filteredY = scipy.signal.savgol_filter(
                self._sweepY,
                SavitzkyGolay_pnts,
                SavitzkyGolay_poly,
                axis=0,
                mode="nearest",
            )
        else:
            self._filteredY = self.sweepY

        self._filteredDeriv = np.diff(self._filteredY, axis=0)

        # filter the derivative
        if medianFilter > 0:
            if not medianFilter % 2:
                medianFilter += 1
                print(
                    f"Please use an odd value for the median filter, set medianFilter: {medianFilter}"
                )
            medianFilter = int(medianFilter)
            self._filteredDeriv = scipy.signal.medfilt2d(
                self._filteredDeriv, [medianFilter, 1]
            )
        elif SavitzkyGolay_pnts > 0:
            self._filteredDeriv = scipy.signal.savgol_filter(
                self._filteredDeriv,
                SavitzkyGolay_pnts,
                SavitzkyGolay_poly,
                axis=0,
                mode="nearest",
            )
        else:
            # self._filteredDeriv = self.filteredDeriv
            pass

        # mV/ms
        dataPointsPerMs = self.dataPointsPerMs
        self._filteredDeriv = self._filteredDeriv * dataPointsPerMs  # / 1000

        # insert an initial point (rw) so it is the same length as raw data in abf.sweepY
        # three options (concatenate, insert, vstack), could only get vstack working
        rowOfZeros = np.zeros(self.numSweeps)

        # logger.info(f' dataPointsPerMs:{dataPointsPerMs}')
        # logger.info(f' self.numSweeps:{self.numSweeps}')
        # logger.info(f' rowOfZeros:{rowOfZeros.shape}')
        # logger.info(f' 1 - _filteredDeriv:{self._filteredDeriv.shape}')

        # rowZero = 0
        self._filteredDeriv = np.vstack([rowOfZeros, self._filteredDeriv])

        # logger.info(f'  sweepX:{self.sweepX.shape}')
        # logger.info(f'  sweepY:{self.sweepY.shape}')
        # logger.info(f'  _filteredY:{self._filteredY.shape}')
        # logger.info(f'  2- _filteredDeriv:{self._filteredDeriv.shape}')

    @property
    def sweepY_filtered(self) -> np.ndarray:
        """Get a filtered version of sweepY."""
        if self._filteredY is not None:
            return self._filteredY[:, self.currentSweep]

    @property
    def recordingFrequency(self) -> int:
        """Convenience for dataPointsPerMs, recording frequency in kHz."""
        return self.dataPointsPerMs

    def pnt2Sec_(self, pnt: int) -> float:
        """Convert a point to seconds using dataPointsPerMs.

        Parameters
        ----------
        pnt : int

        Returns
        -------
        float
            The point in seconds (s)
        """
        if pnt is None:
            # return math.isnan(pnt)
            return math.nan
        else:
            return pnt / self.dataPointsPerMs / 1000

    def pnt2Ms_(self, pnt: int) -> float:
        """
        Convert a point to milliseconds (ms) using `self.dataPointsPerMs`

        Parameters
        ----------
        pnt : int

        Returns
        -------
        float
            The point in milliseconds (ms)
        """
        return pnt / self.dataPointsPerMs

    def ms2Pnt_(self, ms: float) -> int:
        """
        Convert milliseconds (ms) to point in recording using `self.dataPointsPerMs`

        Parameters
        ----------
        ms : float
            The ms into the recording

        Returns
        -------
        int
            The point in the recording corresponding to ms
        """
        theRet = ms * self.dataPointsPerMs
        theRet = round(theRet)
        return theRet

    def getEpochTable(self, sweep: int):
        """Only proper abf files will have an epoch table.

        TODO: Make all file loders have an epoch table.
            Make API so derived file loaders can create their own
        """
        if self._epochTableList is not None:
            return self._epochTableList[sweep]
        else:
            return None

    @property
    def numEpochs(self) -> Optional[int]:
        """Get the number of epochs.

        Epochs are mostly for pClamp abf files. We are assuming each sweep has the same namber of epochs.
        """
        if self._epochTableList is not None:
            return self._epochTableList[0].numEpochs()

    def _checkLoadedData(self):
        # TODO: check all the member vraiables are correct
        # set error if they are not
        pass
    
    def setLoadedData(
        self,
        sweepX: np.ndarray,
        sweepY: np.ndarray,
        sweepC: Optional[np.ndarray] = None,
        recordingMode: recordingModes = recordingModes.iclamp,
        xLabel: str = "",
        yLabel: str = "",
    ):
        """Derived classes call this function once the data is loaded in loadFile().

        Parameters
        ----------
        sweepX : np.ndarray
            Time values
        sweepY : np.ndarray
            Recording values, mV or pA
        sweepC : np.ndarray
            (optional) DAC stimulus, pA or mV
        recordingMode : recordingModes
            (optional) Defaults to recordingModes.iclamp)
        xLabel : str
            (optional) str for x-axis label
        yLabel : str
            (optional) str for y-axis label

        Notes
        -----
        - Number of sweeps: sweepY.shape[1]
        - Sweep Length (sec): sweepX[-1,0]
        - Data Points Per Millisecond: 1 / ((sweepX[1,0] - sweepX[0,0]) * 1000)
        """
        self._sweepX = sweepX
        self._sweepY = sweepY
        self._sweepC = sweepC

        self._numSweeps: int = self._sweepY.shape[1]
        self._sweepList: List[int] = list(range(self._numSweeps))

        self._sweepLengthSec: float = self._sweepX[-1, 0]  # from 0 to last sample point

        dtSeconds = self._sweepX[1, 0] - self._sweepX[0, 0]  # seconds per sample
        dtSeconds = float(dtSeconds)
        dtMilliseconds = dtSeconds * 1000
        _dataPointsPerMs = int(1 / dtMilliseconds)
        # logger.info(f'dtSeconds:{dtSeconds} dtMilliseconds:{dtMilliseconds} _dataPointsPerMs:{_dataPointsPerMs}')
        self._dataPointsPerMs: int = _dataPointsPerMs

        self._recordingMode: recordingModes = recordingMode
        self._sweepLabelX: str = xLabel
        self._sweepLabelY: str = yLabel

if __name__ == "__main__":
    d = getFileLoaders()
    # for k,v in d.items():
    #     print(k,v)



from typing import Union, Optional
from pprint import pprint

import pandas as pd

import pyabf

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


class epochTable:
    """Load epoch/stimulation from abf file using 'sweepEpochs'.

    Values in epoch table are per sweep!
    """

    def __init__(self, abf: pyabf.ABF):
        self._epochList = []

        dataPointsPerMs = abf.dataPointsPerMs
        """To convert point to seconds"""

        try:
            _tmp = abf.sweepEpochs.p1s
        except AttributeError as e:
            logger.error(e)
            return

        # sweepEpochs is type "pyabf.waveform.EpochSweepWaveform"
        for epochIdx, p1 in enumerate(abf.sweepEpochs.p1s):
            p2 = abf.sweepEpochs.p2s[epochIdx]  # stop point of each pulse
            epochLevel = abf.sweepEpochs.levels[epochIdx]
            epochType = abf.sweepEpochs.types[epochIdx]
            pulseWidth = abf.sweepEpochs.pulseWidths[epochIdx]
            pulsePeriod = abf.sweepEpochs.pulsePeriods[epochIdx]
            digitalState = abf.sweepEpochs.pulsePeriods[epochIdx]
            ##print(f"epoch index {epochIdx}: at point {p1} there is a {epochType} to level {epochLevel}")

            p1_sec = p1 / abf.dataPointsPerMs / 1000
            p2_sec = p2 / abf.dataPointsPerMs / 1000

            epochDict = {
                "sweepNumber": abf.sweepNumber,
                "index": epochIdx,
                "type": epochType,
                "startPoint": p1,  # point the epoch starts
                "stopPoint": p2,  # point the epoch ends
                "startSec": p1_sec,
                "stopSec": p2_sec,
                "durSec": p2_sec - p1_sec,
                "level": epochLevel,
                "pulseWidth": pulseWidth,
                "pulsePeriod": pulsePeriod,
                "digitalState": digitalState,  # list of 0/1 for 8x digital states
            }
            self._epochList.append(epochDict)

    def getEpochList(self, asDataFrame: bool = False):
        if asDataFrame:
            return pd.DataFrame(self._epochList)
        else:
            return self._epochList

    def findEpoch(self, pnt: int) -> Optional[int]:
        """Return epoch index for a point in recording.

        Stop points are always the same as next epoch start point.
        Be sure to use '<' like pnt<stopPnt to get epoch index correct.

        If not found, return None

        Parameters
        ----------
        pnt : int
            Point index into recording (within a sweep)
        """
        for epochIdx, epoch in enumerate(self._epochList):
            startPoint = epoch["startPoint"]
            stopPoint = epoch["stopPoint"]
            if pnt >= startPoint and pnt < stopPoint:
                return epochIdx
        #
        # return None

    def getLevel(self, epoch):
        """Given an epoch number return the 'level'"""
        return self._epochList[epoch]["level"]

    def getStartSec(self, epoch):
        """Given an epoch number return the 'startSec'"""
        return self._epochList[epoch]["startSec"]

    def getStartSecs(self):
        """Return all epoch start times."""
        startSecs = [epoch["startSec"] for epoch in self._epochList]
        return startSecs

    def getEpochLines(self, yMin=0, yMax=1):
        x = [float("nan")] * (self.numEpochs() * 3)
        y = [float("nan")] * (self.numEpochs() * 3)

        for epoch in range(self.numEpochs()):
            idx = epoch * 3
            x[idx] = self._epochList[epoch]["startSec"]
            x[idx + 1] = self._epochList[epoch]["startSec"]
            x[idx + 2] = float("nan")

            y[idx] = yMin
            y[idx + 1] = yMax
            y[idx + 2] = float("nan")

        return x, y

    def numEpochs(self):
        # print('qqq:', self._epochList)
        return len(self._epochList)


if __name__ == "__main__":
    import sanpy

    # path = '/Users/cudmore/Sites/SanPy/data/19114000.abf'
    path = "/Users/cudmore/data/theanne-griffith/07.28.21/2021_07_28_0001.abf"

    ba = sanpy.bAnalysis(path)
    print(ba)

    sweep = 13

    et = ba.getEpochTable(sweep)
    df = et.getEpochList(asDataFrame=True)
    pprint(df)

    testPnt = 1280
    epochIndex = et.findEpoch(testPnt)
    print(f"testPnt:{testPnt} is in epoch index: {epochIndex}")

    print("getStartSecs:", et.getStartSecs())
    print("getEpochLines:", et.getEpochLines())



from .fileLoader_abf import fileLoader_abf
from .fileLoader_atf import fileLoader_atf
from .fileLoader_csv import fileLoader_csv
from .fileLoader_tif import fileLoader_tif
from .fileLoader_base import fileLoader_base
from .fileLoader_base import recordingModes
from .fileLoader_base import getFileLoaders

from .epochTable import epochTable



"""sanpy fft backend.

TODO:
	- get rid of (spikeDetect, _getHalfWidths, _throwOutRefractory)
		- these are used to spike detect model data
		- try and use main bAnalysis_
"""

import math, time, os
from math import exp
from functools import partial
from tokenize import blank_re

import numpy as np
import pandas as pd

from scipy.signal import lfilter, freqz
import scipy.signal

import matplotlib as mpl
import matplotlib.pyplot as plt

import sanpy

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


def butter_sos(cutOff, fs, order=50, passType="lowpass"):
    """
    passType (str): ('lowpass', 'highpass')
    """

    """
	nyq = fs * 0.5
	normal_cutoff = cutOff / nyq
	"""

    logger.info(
        f"Making butter file with order:{order} cutOff:{cutOff} passType:{passType} fs:{fs}"
    )

    # sos = butter(order, normal_cutoff, btype='low', analog=False, output='sos')
    # sos = scipy.signal.butter(N=order, Wn=normal_cutoff, btype=passType, analog=False, fs=fs, output='sos')
    sos = scipy.signal.butter(
        N=order, Wn=cutOff, btype=passType, analog=False, fs=fs, output="sos"
    )
    return sos


def spikeDetect(t, dataFiltered, dataThreshold2, startPoint=None, stopPoint=None):
    verbose = False
    # dataThreshold2 = -59.81

    g_backupFraction = 0.3  # 0.5 or 0.3

    logger.info(
        f"dataThreshold2:{dataThreshold2} startPoint:{startPoint} stopPoint:{stopPoint}"
    )

    #
    # detect spikes, dataThreshold2 changes as we zoom
    Is = np.where(dataFiltered > dataThreshold2)[0]
    Is = np.concatenate(([0], Is))
    Ds = Is[:-1] - Is[1:] + 1
    spikePoints = Is[np.where(Ds)[0] + 1]
    spikeSeconds = t[spikePoints]

    #
    # Throw out spikes that are within refractory (ms) of previous
    spikePoints = _throwOutRefractory(spikePoints, refractory_ms=100)
    spikeSeconds = t[spikePoints]

    #
    # remove spikes outside of start/stop
    goodSpikes = []
    if startPoint is not None and stopPoint is not None:
        if verbose:
            print(f"  Removing spikes outside point range {startPoint} ... {stopPoint}")
        # print(f'    {t[startPoint]} ... {t[stopPoint]} (s)')
        for idx, spikePoint in enumerate(spikePoints):
            if spikePoint < startPoint or spikePoint > stopPoint:
                pass
            else:
                goodSpikes.append(spikePoint)
                if verbose:
                    print(f"    appending spike {idx} {t[spikePoint]}(s)")
    #
    spikePoints = goodSpikes
    spikeSeconds = t[spikePoints]

    #
    # only accept rising phase
    windowPoints = 40  # assuming 10 kHz -->> 4 ms
    goodSpikes = []
    for idx, spikePoint in enumerate(spikePoints):
        preClip = dataFiltered[spikePoint - windowPoints : spikePoint - 1]
        postClip = dataFiltered[spikePoint + 1 : spikePoint + windowPoints]
        preMean = np.nanmean(preClip)
        postMean = np.nanmean(postClip)
        if preMean < postMean:
            goodSpikes.append(spikePoint)
        else:
            pass
            # print(f'  rejected spike {idx}, at point {spikePoint}')
    #
    spikePoints = goodSpikes
    spikeSeconds = t[spikePoints]

    #
    # throw out above take-off-potential of APs (-30 mV)
    apTakeOff = -30
    goodSpikes = []
    for idx, spikePoint in enumerate(spikePoints):
        if dataFiltered[spikePoint] < apTakeOff:
            goodSpikes.append(spikePoint)
    #
    spikePoints = goodSpikes
    spikeSeconds = t[spikePoints]

    #
    # find peak for eack spike
    # at 500 pnts, assuming spikes are slower than 20 Hz
    peakWindow_pnts = 500
    peakPoints = []
    peakVals = []
    for idx, spikePoint in enumerate(spikePoints):
        peakPnt = np.argmax(dataFiltered[spikePoint : spikePoint + peakWindow_pnts])
        peakPnt += spikePoint
        peakVal = np.max(dataFiltered[spikePoint : spikePoint + peakWindow_pnts])
        #
        peakPoints.append(peakPnt)
        peakVals.append(peakVal)

    #
    # backup each spike until pre/post is not changing
    # uses g_backupFraction to get percent of dv/dt at spike veruus each backup window
    backupWindow_pnts = 30
    maxSteps = 30  # 30 steps at 20 pnts (2ms) per step gives 60 ms
    backupSpikes = []
    for idx, spikePoint in enumerate(spikePoints):
        # print(f'=== backup spike:{idx} at {t[spikePoint]}(s)')
        foundBackupPoint = None
        for step in range(maxSteps):
            tmpPnt = spikePoint - (step * backupWindow_pnts)
            tmpSeconds = t[tmpPnt]
            preClip = dataFiltered[
                tmpPnt - 1 - backupWindow_pnts : tmpPnt - 1
            ]  # reversed
            postClip = dataFiltered[tmpPnt + 1 : tmpPnt + 1 + backupWindow_pnts]
            preMean = np.nanmean(preClip)
            postMean = np.nanmean(postClip)
            diffMean = postMean - preMean
            if step == 0:
                initialDiff = diffMean
            # print(f'  spike {idx} step:{step} tmpSeconds:{round(tmpSeconds,4)} initialDiff:{round(initialDiff,4)} diff:{round(diffMean,3)}')
            # if diffMean is 1/2 initial AP slope then accept
            if diffMean < (initialDiff * g_backupFraction):
                # stop
                if foundBackupPoint is None:
                    foundBackupPoint = tmpPnt
                # break
        #
        if foundBackupPoint is not None:
            backupSpikes.append(foundBackupPoint)
        else:
            # needed to keep spike parity
            logger.warning(
                f"Did not find backupSpike for spike {idx} at {t[spikePoint]}(s)"
            )
            backupSpikes.append(spikePoint)

    #
    # use backupSpikes (points) to get each spike amplitude
    spikeAmps = []
    for idx, backupSpike in enumerate(backupSpikes):
        if backupSpikes == np.nan or math.isnan(backupSpike):
            continue
        # print('backupSpike:', backupSpike)
        footVal = dataFiltered[backupSpike]
        peakVal = peakVals[idx]
        spikeAmp = peakVal - footVal
        spikeAmps.append(spikeAmp)

    # TODO: use foot and peak to get real half/width
    theseWidths = [10, 20, 50, 80, 90]
    window_ms = 100
    spikeDictList = _getHalfWidths(
        t,
        dataFiltered,
        backupSpikes,
        peakPoints,
        theseWidths=theseWidths,
        window_ms=window_ms,
    )
    # for idx, spikeDict in enumerate(spikeDictList):
    # 	print(idx, spikeDict)

    #
    # get estimate of duration
    # for each spike, find next downward crossing (starting at peak)
    """
	minDurationPoints = 10
	windowPoints = 1000  # 100 ms
	fallingPoints = []
	for idx, spikePoint in enumerate(spikePoints):
		peakPoint = peakPoints[idx]
		backupPoint = backupSpikes[idx]
		if backupPoint == np.nan or math.isnan(backupPoint):
			continue
		spikeSecond = spikeSeconds[idx]
		# Not informative because we are getting spikes from absolute threshold
		thisThreshold = dataFiltered[backupPoint]
		thisPeak = dataFiltered[peakPoint]
		halfHeight = thisThreshold + (thisPeak - thisThreshold) / 2
		startPoint = peakPoint #+ 50
		postClip = dataFiltered[startPoint:startPoint+windowPoints]
		tmpFallingPoints = np.where(postClip<halfHeight)[0]
		if len(tmpFallingPoints) > 0:
			fallingPoint = startPoint + tmpFallingPoints[0]
			# TODO: check if falling point is AFTER next spike
			duration = fallingPoint - spikePoint
			if duration > minDurationPoints:
				fallingPoints.append(fallingPoint)
			else:
				print(f'  reject spike {idx} at {spikeSecond} (s), duration is {duration} points, minDurationPoints:{minDurationPoints}')
				#fallingPoints.append(fallingPoint)
		else:
			print(f'    did not find falling pnt for spike {idx} at {spikeSecond}(s) point {spikePoint}, assume it is longer than windowPoints')
			pass
			#fallingPoints.append(np.nan)
	"""

    # TODO: Package all results into a dictionary
    spikeDictList = [{}] * len(spikePoints)
    for idx, spikePoint in enumerate(spikePoints):
        """
        spikeSecond = spikeSeconds[idx]
        peakPoint = peakPoints[idx]
        peakVal = peakVals[idx]
        fallingPoint = fallingPoints[idx]  # get rid of this
        backupSpike = backupSpikes[idx]  # get rid of this
        spikeAmp = spikeAmps[idx]  # get rid of this
        """
        spikeDictList[idx]["spikePoint"] = spikePoint
        spikeDictList[idx]["spikeSecond"] = spikeSeconds[idx]
        spikeDictList[idx]["peakPoint"] = peakPoints[idx]
        spikeDictList[idx]["peakVal"] = peakVals[idx]
        # spikeDictList[idx]['fallingPoint'] = fallingPoints[idx]
        spikeDictList[idx]["backupSpike"] = backupSpikes[idx]
        spikeDictList[idx]["spikeAmp"] = spikeAmps[idx]
    #
    spikeSeconds = t[spikePoints]
    # return spikeSeconds, spikePoints, peakPoints, peakVals, fallingPoints, backupSpikes, spikeAmps
    # removved fallingPoints
    return (
        spikeDictList,
        spikeSeconds,
        spikePoints,
        peakPoints,
        peakVals,
        backupSpikes,
        spikeAmps,
    )


def _getHalfWidths(
    t, v, spikePoints, peakPoints, theseWidths=[10, 20, 50, 80, 90], window_ms=50
):
    """Get half widths.

    Args:
            t (ndarray): Time
            v (ndaray): Recording (usually curent clamp Vm)
            spikePoints (list of int): List of spike threshold crossings.
                                                    Usually the back-up version
            peakPoints (list of int):
            theseWidths (list of int): Specifies full-width-half maximal to calculate
            window_ms (int): Number of ms to look after the peak for downward 1/2 height crossing

    Returns:
            List of dict, one elemennt per spike
    """
    logger.info(
        f"theseWidths:{theseWidths} window_ms:{window_ms} spikePoints:{len(spikePoints)} peakPoints:{len(peakPoints)}"
    )

    pointsPerMs = 10  # ToDo: generalize this
    window_pnts = window_ms * pointsPerMs

    spikeDictList = [{}] * len(spikePoints)  # an empy list of dict with proper size
    for idx, spikePoint in enumerate(spikePoints):
        # print(f'Spike {idx} pnt:{spikePoint}')

        # each spike has a list of dict for each width result
        spikeDictList[idx]["widths"] = []

        # check each pre/post pnt is before next spike
        nextSpikePnt = None
        if idx < len(spikePoints) - 1:
            nextSpikePnt = spikePoints[idx + 1]

        peakPoint = peakPoints[idx]
        try:
            vSpike = v[spikePoint]
        except IndexError as e:
            logger.error(f"spikePoint:{spikePoint} {type(spikePoint)}")
        vPeak = v[peakPoint]
        height = vPeak - vSpike
        preStartPnt = spikePoint + 1
        preClip = v[preStartPnt : peakPoint - 1]
        postStartPnt = peakPoint + 1
        postClip = v[postStartPnt : peakPoint + window_pnts]
        for width in theseWidths:
            thisHeight = vSpike + (height / width)  # search for a percent of height
            prePnt = np.where(preClip >= thisHeight)[0]
            if len(prePnt) > 0:
                prePnt = preStartPnt + prePnt[0]
            else:
                # print(f'  Error: Spike {idx} "prePnt" width:{width} vSpike:{vSpike} height:{height} thisHeight:{thisHeight}')
                prePnt = None
            postPnt = np.where(postClip < thisHeight)[0]
            if len(postPnt) > 0:
                postPnt = postStartPnt + postPnt[0]
            else:
                # print(f'  Error: Spike {idx} "postPnt" width:{width} vSpike:{vSpike} height:{height} thisHeight:{thisHeight}')
                postPnt = None
            widthMs = None
            if prePnt is not None and postPnt is not None:
                widthPnts = postPnt - prePnt
                widthMs = widthPnts / pointsPerMs
                # print(f'  width:{width} tPre:{t[prePnt]} tPost:{t[postPnt]} widthMs:{widthMs}')
                if nextSpikePnt is not None and prePnt >= nextSpikePnt:
                    print(
                        f"  Error: Spike {idx} widthMs:{widthMs} prePnt:{prePnt} is after nextSpikePnt:{nextSpikePnt}"
                    )
                if nextSpikePnt is not None and postPnt >= nextSpikePnt:
                    print(
                        f"  Error: Spike {idx} widthMs:{widthMs} postPnt:{postPnt} is after nextSpikePnt:{nextSpikePnt}"
                    )

            # put into dict
            widthDict = {
                "halfHeight": width,
                "risingPnt": prePnt,
                #'risingVal': defaultVal,
                "fallingPnt": postPnt,
                #'fallingVal': defaultVal,
                #'widthPnts': None,
                "widthMs": widthMs,
            }
            spikeDictList[idx]["widths_" + str(width)] = widthMs
            spikeDictList[idx]["widths"].append(widthDict)
    #
    return spikeDictList


def _throwOutRefractory(spikePoints, refractory_ms=100):
    """
    spikePoints: spike times to consider
    refractory_ms:
    """
    dataPointsPerMs = 10

    before = len(spikePoints)

    # if there are doubles, throw-out the second one
    # refractory_ms = 20 #10 # remove spike [i] if it occurs within refractory_ms of spike [i-1]
    lastGood = 0  # first spike [0] will always be good, there is no spike [i-1]
    for i in range(len(spikePoints)):
        if i == 0:
            # first spike is always good
            continue
        dPoints = spikePoints[i] - spikePoints[lastGood]
        if dPoints < dataPointsPerMs * refractory_ms:
            # remove spike time [i]
            spikePoints[i] = 0
        else:
            # spike time [i] was good
            lastGood = i
    # regenerate spikeTimes0 by throwing out any spike time that does not pass 'if spikeTime'
    # spikeTimes[i] that were set to 0 above (they were too close to the previous spike)
    # will not pass 'if spikeTime', as 'if 0' evaluates to False
    # if goodSpikeErrors is not None:
    # 	goodSpikeErrors = [goodSpikeErrors[idx] for idx, spikeTime in enumerate(spikeTimes0) if spikeTime]
    spikePoints = [spikePoint for spikePoint in spikePoints if spikePoint]

    # TODO: put back in and log if detection ['verbose']
    after = len(spikePoints)
    logger.info(f"From {before} to {after} spikes with refractory_ms:{refractory_ms}")

    return spikePoints


def getKernel(type="sumExp", amp=5, tau1=30, tau2=70):
    """Get a kernel for convolution with a spike train."""
    N = 500  # pnts
    t = [x for x in range(N)]
    y = t

    if type == "sumExp":
        for i in t:
            y[i] = -amp * (exp(-t[i] / tau1) - (exp(-t[i] / tau2)))
    #
    return y


def getSpikeTrain(numSeconds=1, fs=10000, spikeFreq=3, amp=10, noiseAmp=10):
    """Get a spike train at given frequency.

    Arguments:
            numSeconds (int): Total number of seconds
            fs (int): Sampling frequency, 10000 for 10 kH
            spikeFreq (int): Frequency of events in spike train, e.g. simulated EPSPs
            amp (float): Amplitude of sum exponential kernel (getKernel)
    """
    n = int(numSeconds * fs)  # total number of samples
    numSpikes = int(numSeconds * spikeFreq)
    spikeTrain = np.zeros(n)
    start = fs / spikeFreq
    spikeTimes = np.linspace(start, n, numSpikes, endpoint=False)
    for idx, spike in enumerate(spikeTimes):
        # print(idx, spike)
        spike = int(spike)
        spikeTrain[spike] = 1

    expKernel = getKernel(amp=amp)
    epspTrain = scipy.signal.convolve(spikeTrain, expKernel, mode="same")
    # shift to -60 mV
    epspTrain -= 60

    # add noise
    if noiseAmp == 0:
        pass
    else:
        # noise_power = 0.001 * fs / noiseAmp
        # epspTrain += np.random.normal(scale=np.sqrt(noise_power), size=epspTrain.shape)
        epspTrain += np.random.normal(scale=noiseAmp, size=epspTrain.shape)

    #
    t = np.linspace(0, numSeconds, n, endpoint=True)
    #
    return t, spikeTrain, epspTrain


class fftFromRecording:
    def __init__(
        self,
        path=None,
        ba=None,
        sweepNumber=0,
        startSec=None,
        stopSec=None,
        doPlot=True,
    ):
        """fft backend.

        Args:
                path (str) if specified, load abf from path
                ba (bAnalysis) if specified use that
        """

        if ba is not None:
            self.ba = ba
        elif path is not None:
            self.ba = sanpy.bAnalysis(path)
        else:
            logger.error(f"Expecting either a path or existing bAnalysis object")

        # self.ba.setSweep(sweepNumber)
        self.sweepNumber = sweepNumber

        if startSec is None or stopSec is None:
            self.startSec = 0
            self.stopSec = self.ba.recordingDur
        else:
            self.startSec = startSec
            self.stopSec = stopSec

        self.plotRawAxis = True

        self.doButterFilter = True
        self.butterOrder = 70
        self.butterCutoff = [0.7, 10]  # [low freq, high freq] for bandpass

        self._resultsDictList = []
        self._resultsDictList2 = []
        self._resultStr = ""

        self.isModel = False

        self._store_ba = None  # allow switching between model and self.ba

        if self.ba is not None:
            self.fs = self.ba.recordingFrequency * 1000
        else:
            self.fs = None
        self.psdWindowStr = "Hanning"  # mpl.mlab.window_hanning

        # fft Freq (Hz) resolution is fs/nfft --> resolution 0.2 Hz = 10000/50000
        self.nfft = 50000  # 512 * 100

        self.signalHz = None  # assign when using fake data
        self.maxPlotHz = 10  # limit x-axis frequenccy

        self.medianFilterPnts = 0

        self.dataLine = None
        self.dataFilteredLine = None
        self.spikesLine = None
        self.peaksLine = None
        self.dataMeanLine = None
        self.thresholdLine2 = None
        self.thresholdLine3 = None

        # self.fig = mpl.figure.Figure(constrained_layout=True)
        self.fig = plt.figure(constrained_layout=True)

        if self.plotRawAxis:
            numRow = 3
        else:
            numRow = 2

        gs = self.fig.add_gridspec(numRow, 3)

        if self.plotRawAxis:
            self.rawAxes = self.fig.add_subplot(gs[0, :])
            self.rawZoomAxes = self.fig.add_subplot(gs[1, :])
            # 3rd row has 3 subplots
            self.psdAxes = self.fig.add_subplot(gs[2, 0])
            self.spectrogramAxes = self.fig.add_subplot(gs[2, -2])
        else:
            self.rawAxes = None
            self.rawZoomAxes = self.fig.add_subplot(gs[0, :])
            # 3rd row has 3 subplots
            self.psdAxes = self.fig.add_subplot(gs[1, 0])
            self.spectrogramAxes = self.fig.figure.add_subplot(gs[1, -2])

        if doPlot:
            self.replot_fft2(sweepNumber=self.sweepNumber)

    def getFigFileName(self):
        filename = self.ba.getFileName()
        filename = os.path.splitext(filename)[0]
        sweep = self.ba.currentSweep
        figFileName = f"{filename}_sweep_{sweep}.png"
        return figFileName

    def saveFig(self, folderPath=None):
        """Save the current figure."""

        # TODO: if folderPath then check it exists

        figFileName = self.getFigFileName()
        figFilePath = os.path.join(folderPath, figFileName)

        logger.info(f"Saving figure to: {figFilePath}")

        self.fig.savefig(figFilePath, dpi=300)

    def getStartStop(self):
        # logger.warning('remember to set start/stop')
        # startSec = 0
        # stopSec = self.ba.recordingDur
        return self.startSec, self.stopSec

    def linearDetrend(self, x):
        # print('myDetrend() x:', x.shape)
        y = plt.mlab.detrend_linear(x)
        # y = plt.mlab.detrend_mean(y)
        return y

    def replot_fft2(self, sweepNumber=0):
        start = time.time()

        if self.ba is None:
            return

        self.ba.setSweep(sweepNumber)

        # if self.sweepNumber == 'All':
        # 	logger.warning(f'fft plugin can only show one sweep, received sweepNumber:{self.sweepNumber}')
        # 	return

        # logger.info(f'using ba: {self.ba}')
        startSec, stopSec = self.getStartStop()
        if (
            startSec is None
            or stopSec is None
            or math.isnan(startSec)
            or math.isnan(stopSec)
        ):
            logger.info(f"Resetting start/stop seconds to max")
            startSec = 0
            stopSec = self.ba.recordingDur  # self.ba.sweepX()[-1]
        logger.info(f"Using start(s):{startSec} stop(s):{stopSec}")
        self.lastLeft = round(startSec * 1000 * self.ba.dataPointsPerMs)
        self.lastRight = round(stopSec * 1000 * self.ba.dataPointsPerMs)

        leftPoint = self.lastLeft
        rightPoint = self.lastRight

        # sweepY = self.getSweep('y')
        sweepY = self.ba.sweepY
        if leftPoint < 0:
            leftPoint = 0
        if rightPoint > sweepY.shape[0]:
            rightPoint = sweepY.shape[0]

        y = sweepY[leftPoint:rightPoint]

        medianPnts = self.medianFilterPnts  # 50
        if medianPnts > 0:
            logger.info(f"  Median filter with pnts={medianPnts}")
            yFiltered = scipy.ndimage.median_filter(y, medianPnts)
        else:
            logger.info("  No median filter")
            yFiltered = y

        # logger.info(f'Fetching sweepX with sweepNumber: {self.sweepNumber}')

        # sweepX = self.getSweep('x')
        sweepX = self.ba.fileLoader.sweepX

        # logger.info(f'  sweepX: {sweepX.shape}')
        # logger.info(f'  leftSec:{sweepX[leftPoint]} rightSec:{sweepX[rightPoint-1]}')
        t = sweepX[leftPoint:rightPoint]

        dt = t[1] - t[0]  # 0.0001
        fs = round(1 / dt)  # samples per second
        nfft = self.nfft  # The number of data points used in each block for the FFT
        # print(f'  fs:{fs} nfft:{nfft}')

        # 20220329 hijacking raw plot to show stim
        if self.plotRawAxis:
            sweepC = self.ba.sweepC

            self.rawAxes.clear()
            # self.rawAxes.plot(sweepX, sweepY, '-', linewidth=1)
            self.rawAxes.plot(sweepX, sweepC, "-", linewidth=1)
            # draw a rectangle to show left/right time selecction
            """
			yMin = np.nanmin(sweepY)
			yMax = np.nanmax(sweepY)
			xRect = startSec
			yRect = yMin
			wRect = stopSec-startSec
			hRect = yMax-yMin
			rect1 = mpl.patches.Rectangle((xRect,yRect), wRect, hRect, color='gray')
			self.rawAxes.add_patch(rect1)
			"""
            # self.rawAxes.set_xlim([sweepX[0], sweepX[-1]])
            self.rawAxes.set_xlim([self.startSec, self.stopSec])

        #
        # replot the clip +/- std we analyzed, after detrend
        yDetrend = self.linearDetrend(yFiltered)

        #
        if self.doButterFilter:
            # self.butterCutoff = 10  # Hz
            # self.butterOrder = 70
            # self.sos = butter_sos(self.butterCutoff, self.fs, order=self.butterOrder, passType='lowpass')
            # self.butterCutoff = [0.7, 10]
            # logger.info(f'  butterOrder:{self.butterOrder} butterCutoff:{self.butterCutoff}')
            self.sos = butter_sos(
                self.butterCutoff, self.fs, order=self.butterOrder, passType="bandpass"
            )
            # filtfilt should remove phase delay in time. A forward-backward digital filter using cascaded second-order sections.
            yFiltered_Butter = scipy.signal.sosfiltfilt(self.sos, yDetrend, axis=0)
            # print('yDetrend:', yDetrend.shape, np.nanmin(yDetrend), np.nanmax(yDetrend))
            # print('yFiltered_Butter:', yFiltered_Butter.shape, np.nanmin(yFiltered_Butter), np.nanmax(yFiltered_Butter))

            # we use this for remaining
            yFiltered = yFiltered_Butter

        dataMean = np.nanmean(yDetrend)
        dataStd = np.nanstd(yDetrend)
        dataThreshold2 = dataMean + (2 * dataStd)
        dataThreshold3 = dataMean + (3 * dataStd)

        self.rawZoomAxes.clear()
        self.rawZoomAxes.plot(t, yDetrend, "-", linewidth=1)
        self.rawZoomAxes.axhline(
            dataMean, marker="", color="r", linestyle="--", linewidth=0.5
        )
        self.rawZoomAxes.axhline(
            dataThreshold2, marker="", color="r", linestyle="--", linewidth=0.5
        )
        self.rawZoomAxes.axhline(
            dataThreshold3, marker="", color="r", linestyle="--", linewidth=0.5
        )
        self.rawZoomAxes.set_xlim([t[0], t[-1]])
        #
        if self.doButterFilter:
            self.rawZoomAxes.plot(t, yFiltered, "-r", linewidth=1)

        # save yDetrend, yFiltered as csv
        """
		print('=== FOR FERNANDO')
		import pandas as pd
		print('  t:', t.shape)
		print('  yDetrend:', yDetrend.shape)
		print('  yFiltered:', yFiltered.shape)
		tmpDf = pd.DataFrame(columns=['s', 'yDetrend', 'yFiltered'])
		tmpDf['s'] = t
		tmpDf['yDetrend'] = yDetrend
		tmpDf['yFiltered'] = yFiltered
		print(tmpDf.head())
		dfFile = 'fft-20200707-0000-16sec-31sec.csv'
		print('saving dfFile:', dfFile)
		tmpDf.to_csv(dfFile, index=False)
		print('=== END')
		"""

        # we will still scale a freq plots to [0, self.maxPlotHz]
        if self.doButterFilter:
            minPlotFreq = self.butterCutoff[0]
            maxPlotFreq = self.butterCutoff[1]
        else:
            minPlotFreq = 0
            maxPlotFreq = self.maxPlotHz  # 15

        #
        # spectrogram
        # self.fftAxes.clear()
        nfft2 = int(nfft / 4)
        specSpectrum, specFreqs, spec_t, specIm = self.spectrogramAxes.specgram(
            yFiltered, NFFT=nfft2, Fs=fs, detrend=self.linearDetrend
        )
        """
		print('=== specgram')
		print('  yFiltered:', yFiltered.shape)
		print('  specSpectrum (freq, t):', specSpectrum.shape)  # (25601, 20) is (freq, t)
		print('    ', np.nanmin(specSpectrum), np.nanmax(specSpectrum))
		print('  specFreqs:', specFreqs.shape)
		print('  spec_t:', spec_t.shape)
		print('  specIm:', type(specIm))
		"""
        mask = (specFreqs >= minPlotFreq) & (specFreqs <= maxPlotFreq)
        specSpectrum = specSpectrum[mask, :]  # TRANSPOSE
        """
		print('  2 Transpose specSpectrum (freq, t):', specSpectrum.shape)  # (25601, 20) is (freq, t)
		print('    ', np.nanmin(specSpectrum), np.nanmax(specSpectrum))
		"""

        # careful: I want to eventually scale y-axis to [0, self.maxPlotHz]
        # for now use [minPlotFreq, maxPlotFreq]
        x_min = startSec
        x_max = stopSec
        y_min = maxPlotFreq  # minPlotFreq
        y_max = minPlotFreq  # maxPlotFreq
        extent = [x_min, x_max, y_min, y_max]

        self.spectrogramAxes.clear()
        self.spectrogramAxes.imshow(
            specSpectrum, extent=extent, aspect="auto", cmap="viridis"
        )
        self.spectrogramAxes.invert_yaxis()
        self.spectrogramAxes.set_xlabel("Time (s)")
        self.spectrogramAxes.set_ylabel("Freq (Hz)")

        #
        # matplotlib psd
        if self.psdWindowStr == "Hanning":
            psdWindow = mpl.mlab.window_hanning
        elif self.psdWindowStr == "Blackman":
            psdWindow = np.blackman(nfft)
        else:
            logger.warning(f"psdWindowStr not understood {self.psdWindowStr}")
        """
		print('=== calling mpl psd()')
		print('  nfft:', nfft)
		print('  fs:', fs)
		print('  myDetrend:', myDetrend)
		print('  psdWindowStr:', self.psdWindowStr)
		print('  psdWindow:', psdWindow)
		"""
        # default scale_by_freq=True
        scale_by_freq = True
        Pxx, freqs = self.psdAxes.psd(
            yFiltered,
            marker="",
            linestyle="-",
            NFFT=nfft,
            Fs=fs,
            scale_by_freq=scale_by_freq,
            detrend=self.linearDetrend,
            window=psdWindow,
        )

        #
        # replot matplotlib psd
        pxxLog10 = 10 * np.log10(Pxx)

        # pxxLog10 = Pxx # does this look better ???

        mask = (freqs >= minPlotFreq) & (freqs <= maxPlotFreq)
        pxxLog10 = pxxLog10[mask]  # order matters
        freqsLog10 = freqs[mask]  # order matters
        self.psdAxes.clear()
        self.psdAxes.plot(freqsLog10, pxxLog10, "-", linewidth=1)
        self.psdAxes.set_xlim([0, self.maxPlotHz])  # x-axes is frequency
        self.psdAxes.grid(True)
        # self.psdAxes.set_ylabel('10*log10(Pxx)')
        self.psdAxes.set_ylabel("PSD (dB)")
        self.psdAxes.set_xlabel("Freq (Hz)")

        #
        # get peak frequency from psd, finding max peak with width
        inflection = np.diff(np.sign(np.diff(pxxLog10)))
        peaks = (inflection < 0).nonzero()[0] + 1
        # excception ValueError
        try:
            peak0 = peaks[pxxLog10[peaks].argmax()]
            maxFreq0 = round(freqsLog10[peak0], 3)  # Gives 0.05
            maxPsd0 = round(pxxLog10[peak0], 3)  # Gives 0.05
        except ValueError as e:
            logger.error("BAD PEAK in pxx")
            maxFreq0 = []
            maxPsd0 = []
        # print(f'1 Results: Peak Hz={round(maxFreq0,3)} Amplitude={round(maxPsd0,3)}')

        # add to plot
        self.psdAxes.plot(maxFreq0, maxPsd0, ".r")

        # the absolute maximum in the psd
        maxPsd = np.nanmax(pxxLog10)
        maxPnt = np.argmax(pxxLog10)
        maxFreq = freqsLog10[maxPnt]
        # self.resultsLabel.setText(f'Results: Peak Hz={round(maxFreq,3)} Amplitude={round(maxPsd,3)}')
        # print(f'2 Results: Peak Hz={round(maxFreq,3)} Amplitude={round(maxPsd,3)}')

        #
        # print results
        # (file, startSec, stopSec, max psd amp, max psd freq)
        pStart = round(startSec, 3)
        pStop = round(stopSec, 3)
        pMaxFreq = round(maxFreq, 3)
        pMaxPsd = round(maxPsd, 3)

        """
		printStr = f'Type\tFile\tstartSec\tstopSec\tmaxFreqPsd\tmaxPsd'  #\tmaxFreqFft\tmaxFft'
		#self.appendResultsStr(printStr)
		print('=== FFT results are:')
		print(printStr)
		printStr = f'fftPlugin\t{self.ba.getFileName()}\t{pStart}\t{pStop}\t{pMaxFreq}\t{pMaxPsd}'
		self.appendResultsStr(printStr, maxFreq=pMaxFreq, maxPsd=pMaxPsd, 
								maxFreq0=maxFreq0, maxPsd0=maxPsd0,
								freqs=freqsLog10, psd=pxxLog10)  # last 2 are not used
		print(printStr)
		"""

        # 20220330, rms error of stim
        tmpSweepC = sweepC[leftPoint:rightPoint]
        rms = np.sqrt(np.mean(tmpSweepC**2))

        resultDict = {
            "file": self.ba.getFileName(),
            "sweep": self.ba.currentSweep,
            "startSec": self.getStartStop()[0],
            "stopSec": self.getStartStop()[1],
            "butterFilter": self.doButterFilter,
            "butterOrder": self.butterOrder,
            "lowFreqCutoff": self.butterCutoff[0],
            "highFreqCutoff": self.butterCutoff[1],
            # ultimate max
            "maxFreq": pMaxFreq,
            "maxPSD": pMaxPsd,
            # max in peak
            "maxFreq0": maxFreq0,
            "maxPSD0": maxPsd0,
            # TODO: max as stim freq
            "maxFreq_at_stim": "",
            "maxPSD_at_stim": "",
            "rmsStim": rms,  # rms of injected current
            #'freqs': freqs,
            #'psd': psd,
        }
        self.appendResults2(resultDict)

        # TODO: make a dict of results and append to pandas df

        #
        # plot np fft
        # see: https://www.gw-openscience.org/tutorial05/
        # blackman is supposed to correct for low freq signal ???
        """
		doBlackman = False
		if doBlackman:
			window = np.blackman(yFiltered.size)
			windowed_yFiltered = yFiltered*window
		else:
			windowed_yFiltered = yFiltered

		ft = np.fft.rfft(windowed_yFiltered)
		# not sure to use diff b/w [1]-[0] or fs=10000 ???
		tmpFs = t[1]-t[0]
		fftFreqs = np.fft.rfftfreq(len(windowed_yFiltered), tmpFs) # Get frequency axis from the time axis
		fftMags = abs(ft) # We don't care about the phase information here

		# find max peak with width
		inflection = np.diff(np.sign(np.diff(fftMags)))
		peaks = (inflection < 0).nonzero()[0] + 1
		peak = peaks[fftMags[peaks].argmax()]
		signal_freq = round(fftFreqs[peak],3) # Gives 0.05
		signal_mag = round(fftMags[peak],3) # Gives 0.05
		#printStr = f'FFT\t{self.ba.getFileName()}\t{pStart}\t{pStop}\t{signal_freq}\t{signal_mag}'
		#print(printStr)
		printStr += f'\t{signal_freq}\t{signal_mag}'
		print(printStr)
		self.appendResultsStr(printStr)
		#print(f'  fft signal frequency is:{signal_freq} with mag:{signal_mag}')

		# strip down to min/ax x-plot of freq
		mask = (fftFreqs>=minPlotFreq) & (fftFreqs<=maxPlotFreq)
		fftMags = fftMags[mask] # order matters
		fftFreqs = fftFreqs[mask] # order matters

		fftMags = fftMags[2:-1]
		fftFreqs = fftFreqs[2:-1]

		self.fftAxes.clear()
		self.fftAxes.semilogy(fftFreqs, fftMags, '-', linewidth=1)
		self.fftAxes.set_xlim([minPlotFreq, maxPlotFreq])  # x-axes is frequency
		self.fftAxes.set_xlabel('Freq (Hz)')
		self.fftAxes.set_ylabel('FFT Mag')
		self.fftAxes.grid(True)
		self.fftAxes.semilogy(signal_freq, signal_mag, '.r')
		"""

        stop = time.time()
        # logger.info(f'Took {stop-start} seconds.')

        #
        # self.static_canvas.draw()

        figFileName = self.getFigFileName()
        self.fig.suptitle(figFileName)

    def appendResults2(self, resultDict):
        self._resultsDictList2.append(resultDict)

    def appendResultsStr(
        self, str, maxFreq="", maxPsd="", maxFreq0="", maxPsd0="", freqs="", psd=""
    ):
        self._resultStr += str + "\n"
        resultDict = {
            "file": self.ba.getFileName(),
            "sweep": self.ba.currentSweep,
            "startSec": self.getStartStop()[0],
            "stopSec": self.getStartStop()[1],
            "butterFilter": self.doButterFilter,
            "butterOrder": self.butterOrder,
            "lowFreqCutoff": self.butterCutoff[0],
            "highFreqCutoff": self.butterCutoff[1],
            # ultimate max
            "maxFreq": maxFreq,
            "maxPSD": maxPsd,
            # max in peak
            "maxFreq0": maxFreq0,
            "maxPSD0": maxPsd0,
            #'freqs': freqs,
            #'psd': psd,
        }
        self._resultsDictList.append(resultDict)

    def getResultsDictList(self):
        return self._resultsDictList

    def getResultStr(self):
        return self._resultStr


def run():
    # need to use colinAnalysis.bAnalysis2() to load stim file atf
    # see how far I get without it, all I need from it is
    # 	- (sin, noise) params per sweep
    # 	- if a sweep is blank_re
    # 	- try and scrip these things into a table and just use a regular bAnalysis

    #
    # if we were recorded with a stimulus file abf
    # needed to assign stimulusWaveformFromFile
    # tmpSweepC = ba.abf.sweepC

    from pprint import pprint

    masterDf = pd.DataFrame()
    saveFolder = "/Users/cudmore/Desktop/stoch-plots"

    folderPath = (
        "/Users/cudmore/Library/CloudStorage/Box-Box/data/stoch-res/new20220104"
    )

    from sanpy.bAnalysisStim import bAnalysisStim

    for file in sorted(os.listdir(folderPath)):
        if not file.endswith(".abf"):
            continue

        fullPath = os.path.join(folderPath, file)

        ba = bAnalysisStim(
            fullPath
        )  # bAnalysis that loads my stim abf (with header from stimgen)
        # ba = sanpy.bAnalysis(fullPath)

        if ba.numSweeps < 2:
            print("!!! skipping file:", file)
            continue

        print(ba)

        dfStimFile = ba._stimFileDf
        pprint(dfStimFile)

        sweep = 0
        fftRecording = fftFromRecording(
            ba=ba, sweepNumber=sweep, startSec=5, stopSec=25, doPlot=False
        )

        fftRecording.medianFilterPnts = 5

        for sweep in range(ba.numSweeps):
            print(f"{sweep} {ba}")
            fftRecording.replot_fft2(sweepNumber=sweep)
            fftRecording.saveFig(folderPath=saveFolder)

        # resultsDictList = fftRecording.getResultsDictList()
        resultsDictList = fftRecording._resultsDictList2
        df = pd.DataFrame(resultsDictList)

        # add stim params to each sweep
        df["stimFreq"] = dfStimFile["freq(Hz)"]
        df["noiseAmp"] = dfStimFile["noise amp"]

        masterDf = pd.concat([masterDf, df], ignore_index=True)

    #
    pprint(masterDf)

    summaryFile = os.path.join(saveFolder, "stoch-summary.csv")
    print("saving:", summaryFile)
    masterDf.to_csv(summaryFile)

    # plt.show()


if __name__ == "__main__":
    run()



# 20210129

import os, sys
import numpy as np

import tifffile

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


class bAbfText:
    """
    mimic a pyabf file to load from a text file
    """

    def __init__(self, path=None, theDict=None):
        """
        path is either:
            path to .csv file with columns (time (seconds), vm)
            path to .tif file

        theDict:
            sweepX
            sweepY
        """

        self.path = path

        self.sweepX = None
        self.sweepY = None
        self.dataRate = None
        self.dataPointsPerMs = None
        self.sweepList = [0]

        self.tif = None
        self.tifNorm = None  # not used
        self.tifHeader = None

        if theDict is not None:
            logger.info("bAbfText() from dict")
            self.sweepX = theDict["sweepX"]
            self.sweepY = theDict["sweepY"]
        elif path.endswith(".tif"):
            self.sweepX, self.sweepY, self.tif, self.tifNorm = self._abfFromLineScanTif(
                path
            )
            # self.tifHeader = self._loadLineScanHeader(path)

            # logger.info(f'tif params: {np.min(self.tif)} {np.max(self.tif)} {self.tif.shape} {self.tif.dtype}')
        else:
            logger.info("bAbfText() from path: {path}")
            if not os.path.isfile(path):
                logger.error("Did not find file: {path}")
                return

            tmpNumPy = np.loadtxt(path, skiprows=1, delimiter=",")
            self.sweepX = tmpNumPy[:, 0]
            self.sweepY = tmpNumPy[:, 1]

        #
        # linescan is like
        # 'secondsPerLine': 0.0042494285714285715,
        # 1 / secondsPerLine = 235.32
        # 20220926, if this is '1' second, detection will fail
        #  because self.dataPointsPerMs goes to 0
        secondsPerSample = self.sweepX[1] - self.sweepX[0]
        samplesPerSecond = 1 / secondsPerSample
        self.dataRate = samplesPerSecond  # 235 #10000 # samples per second

        # todo: calculat from dataRate
        secondsPerSample = 1 / samplesPerSecond
        msPerSample = secondsPerSample * 1000
        samplesPerMs = 1 / msPerSample
        self.dataPointsPerMs = samplesPerMs  # 4.255 #1/235*1000,10 # 10 kHz

        self.sweepUnitsX = "todo: fix"
        self.sweepUnitsY = "todo: fix"

        # self.myRectRoi = None
        self.myRectRoiBackground = None
        # self.minLineScan = None

        # done in _abfFromLineScanTif
        # defaultRect = self.defaultTifRoi_background()  # for kymograph
        # self.updateTifRoi_background(defaultRect)

        # defaultRect = self.defaultTifRoi()  # for kymograph
        # self.updateTifRoi(defaultRect)

        #
        # print results
        """
        if self.tifHeader is not None:
            for k,v in self.tifHeader.items():
                print('  ', k, ':', v)

        print('  ', 'bAbfText secondsPerSample:', secondsPerSample, 'seconds/sample')
        print('  ', 'bAbfText self.dataRate:', self.dataRate, 'samples/second')
        print('  ', 'bAbfText self.dataPointsPerMs:', self.dataPointsPerMs)
        """

    def defaultTifRoi_background(self):
        """
        Default ROI for kymograph background

        - height/width about 5 pixels
        - position near the end of the line
        """
        widthTif = self.tif.shape[1]
        heightTif = self.tif.shape[0]
        insetPixels = 10  # from end of line (top)
        roiWidthHeight = 5
        # tifHeightPercent = self.tif.shape[0] * 0.15

        left = int(widthTif / 2)
        top = heightTif - insetPixels
        right = left + roiWidthHeight - 1
        bottom = top - roiWidthHeight + 1

        return [left, top, right, bottom]

    def resetRoi(self):
        """reset the kymograph roi to default"""
        defaultRect = self.defaultTifRoi()
        self.updateTifRoi(defaultRect)
        return defaultRect

    def defaultTifRoi(self):
        """
        A default rectangular ROI for main kymograph analysis.

        TODO: Remove hard coding of 0.15 percent of height

        Return:
            list: [left, top, right, bottom]
        """
        xRoiPos = 0  # startSeconds
        yRoiPos = 0  # pixels
        widthRoi = self.tif.shape[1]
        heightRoi = self.tif.shape[0]
        tifHeightPercent = self.tif.shape[0] * 0.15
        yRoiPos += tifHeightPercent
        heightRoi -= 2 * tifHeightPercent

        left = xRoiPos
        top = yRoiPos + heightRoi
        right = xRoiPos + widthRoi
        bottom = yRoiPos

        left = int(left)
        top = int(top)
        right = int(right)
        bottom = int(bottom)

        return [left, top, right, bottom]

    def getTifRoiBackground(self):
        return self.myRectRoiBackground

    def getTifRoi(self):
        return self.myRectRoi

    def updateTifRoi_background(self, theRect=None):
        if theRect is None:
            theRect = self.defaultTifRoi_background()
        if theRect is not None:
            left = theRect[0]
            top = theRect[1]
            right = theRect[2]
            bottom = theRect[3]

        self.myRectRoiBackground = theRect

        tif = self.tif
        theSum = np.sum(tif[bottom:top, left:right])

        return theSum

    def updateTifRoi(self, theRect=None):
        """
        Recalculate (sweepX, sweepY) based on user specified rect ROI

        Args:
            theRect (list): left, top, right, bottom

        Notes:
            This is doing 'background subtract' on sum of line - sum in background roi
                I don't think this is correct???
                'sum of line' has num pixels in scan
                'sum background roi' has a different number of pixels !!!
        """

        logger.info(theRect)

        self.myRectRoi = theRect

        firstFrameSeconds = 0

        tif = self.tif

        left = 0  # startSec
        top = tif.shape[0]
        right = tif.shape[1]  # stopSec
        bottom = 0
        if theRect is not None:
            # left = theRect[0]
            top = theRect[1]
            # right = theRect[2]
            bottom = theRect[3]

        # number of pixels in subset of line scan
        numPixels = top - bottom + 1

        # theBackground = self.updateTifRoi_background()

        # logger.warning(f'I turned off background subtract .. placement of background is sometimes bad')

        _secondsPerLine = self.tifHeader["secondsPerLine"]
        if _secondsPerLine is None:
            _secondsPerLine = 0.001  # so we get 1 pnt per ms
            logger.warning(
                f"No header, secondsPerLine is None. Using fake _secondsPerLine:{_secondsPerLine}"
            )

        # retains shape of entire tif
        yLineScanSum = [np.nan] * tif.shape[1]
        xLineScanSum = [np.nan] * tif.shape[1]
        for i in range(tif.shape[1]):
            # sum each column and background subtract

            # always assign time
            xLineScanSum[i] = firstFrameSeconds + (_secondsPerLine * i)
            # if self.tifHeader['secondsPerLine'] is None:
            #     # no units on kymograph
            #     _fakeSecondsPerLine = 0.001
            #     #logger.error(f'No header, secondsPerLine is None. Using _fakeSecondsPerLine:{_fakeSecondsPerLine}')
            #     xLineScanSum[i] = firstFrameSeconds + (_fakeSecondsPerLine * i)
            # else:
            #     xLineScanSum[i] = firstFrameSeconds + (self.tifHeader['secondsPerLine'] * i)

            if i < left or i > right:
                continue

            # each line scan
            theSum = np.sum(tif[bottom:top, i])

            # background subtract
            # print(f'   === theSum:{theSum} {type(theSum)} theBackground:{theBackground} {type(theBackground)}')
            # if not isinstance(theSum, np.uint64) or not isinstance(theBackground, np.uint64):
            #    print(f'   === theSum:{theSum} {type(theSum)} theBackground:{theBackground} {type(theBackground)}')

            # logger.warning(f'I turned off background subtract .. placement of background is sometimes bad')
            # theSum -= theBackground

            # sys.exit(1)

            # f / f_0
            theSum /= numPixels  # tif.shape[0] # norm to number of pixels in line
            yLineScanSum[i] = theSum

        xLineScanSum = np.asarray(xLineScanSum)
        yLineScanSum = np.asarray(yLineScanSum)

        # derive simplified f/f0 by dividing by min
        self.minLineScan = np.nanmin(yLineScanSum)

        # 20221003, do not do this, ysweep will be sum/pnts
        # yLineScanSum /= self.minLineScan

        logger.info("updating sweepX and sweepY with new rect ROI")
        self.sweepX = xLineScanSum
        self.sweepY = yLineScanSum

    def setSweep(self, sweep):
        pass

    def _abfFromLineScanTif(self, path, theRect=None):
        """

        Returns:
            xLineScanSum (np.ndarray) e.g. sweepX
            yLineScanSum (np.ndarray) e.g. sweepY
        """
        if not os.path.isfile(path):
            logger.error(f"Did not find file: {path}")
            return None, None, None, None

        tif = tifffile.imread(path)
        logger.info(f"Loaded tif with shape: {tif.shape}")

        if len(tif.shape) == 3:
            tif = tif[:, :, 1]  # assuming image channel is 1

        # we want our tiffs to be short and wide
        # shape[0] is space, shape[1] is time
        if tif.shape[0] < tif.shape[1]:
            # correct shape
            pass
        else:
            tif = np.rot90(tif)  # rotates 90 degrees counter-clockwise

        f0 = tif.mean()
        tifNorm = (tif - f0) / f0

        self.tif = tif

        # assuming this was exported using Olympus software
        # which gives us a .txt file
        self.tifHeader = self._loadLineScanHeader(path)
        # self.tifHeader['shape'] = tif.shape
        # self.tifHeader['secondsPerLine'] = \
        #                self.tifHeader['totalSeconds'] / self.tifHeader['shape'][1]
        # tifHeader['abfPath'] = abfFile

        # updateTifRoi sets sweepX and seepY
        defaultRect = self.defaultTifRoi()
        self.updateTifRoi(defaultRect)

        firstFrameSeconds = 0

        #
        # sum the inensities of each image line scan
        # print('tif.shape:', tif.shape) # like (146, 10000)
        """
        yLineScanSum = [np.nan] * tif.shape[1]
        xLineScanSum = [np.nan] * tif.shape[1]
        for i in range(tif.shape[1]):
            theSum = np.sum(tif[:,i])
            theSum /= tif.shape[0] # norm to number of pixels in line
            yLineScanSum[i] = theSum
            xLineScanSum[i] = firstFrameSeconds + (self.tifHeader['secondsPerLine'] * i)

        xLineScanSum = np.asarray(xLineScanSum)
        yLineScanSum = np.asarray(yLineScanSum)
        """

        # normalize to 0..1
        # 20220114, was this
        # yLineScanSum = self._NormalizeData(yLineScanSum)

        # print('bAbfText xLineScanSum:', xLineScanSum.shape)
        # print('bAbfText yLineScanSum:', yLineScanSum.shape)

        return self.sweepX, self.sweepY, tif, tifNorm

    def _loadLineScanHeader(self, path):
        """Load scale from .txt file saved by olympus export.

        Args:
            path: full path to tif

        returns dict:
            numPixels:
            umLength:
            umPerPixel:
            totalSeconds:
        """

        def _defaultHeader():
            defaultHeader = {
                "tif": path,
                "numLines": self.tif.shape[1],
                "totalSeconds": None,
                "shape": self.tif.shape,
                "secondsPerLine": 0.001,
                #'linesPerSecond': None,  # 1/secondsPerLine
                "numPixels": self.tif.shape[0],
                "umLength": None,
                "umPerPixel": 1,
            }
            return defaultHeader

        # "X Dimension"    "138, 0.0 - 57.176 [um], 0.414 [um/pixel]"
        # "T Dimension"    "1, 0.000 - 35.496 [s], Interval FreeRun"
        # "Image Size(Unit Converted)"    "57.176 [um] * 35500.000 [ms]"
        txtFile = os.path.splitext(path)[0] + ".txt"

        theRet = _defaultHeader()
        # theRet = {'tif': path}
        # theRet['numLines'] = self.tif.shape[1]

        if not os.path.isfile(txtFile):
            logger.warning(f"Did not find file: {txtFile}")
            logger.warning(f"  The Kymograph will have no scale.")
            return theRet

        # to open export .txt file for Olympus FV3000
        with open(txtFile, "r") as fp:
            lines = fp.readlines()
            for line in lines:
                line = line.strip()
                if line.startswith('"X Dimension"'):
                    line = line.replace('"', "")
                    line = line.replace(",", "")
                    # print('loadLineScanHeader:', line)
                    # 2 number of pixels in line
                    # 5 um length of line
                    # 7 um/pixel
                    splitLine = line.split()
                    for idx, split in enumerate(splitLine):
                        # print('  ', idx, split)
                        if idx == 2:
                            numPixels = int(split)
                            theRet["numPixels"] = numPixels
                        elif idx == 5:
                            umLength = float(split)
                            theRet["umLength"] = umLength
                        elif idx == 7:
                            umPerPixel = float(split)
                            theRet["umPerPixel"] = umPerPixel

                elif line.startswith('"T Dimension"'):
                    line = line.replace('"', "")
                    line = line.replace(",", "")
                    # print('loadLineScanHeader:', line)
                    # 5 total duration of image acquisition (seconds)
                    splitLine = line.split()
                    for idx, split in enumerate(splitLine):
                        # print('  ', idx, split)
                        if idx == 5:
                            totalSeconds = float(split)
                            theRet["totalSeconds"] = totalSeconds

                            # theRet['secondsPerLine'] =

                # elif line.startswith('"Image Size(Unit Converted)"'):
                #    print('loadLineScanHeader:', line)

        # theRet['shape'] = self.tif.shape
        theRet["secondsPerLine"] = theRet["totalSeconds"] / theRet["shape"][1]
        # theRet['linesPerSecond'] = 1 / theRet['secondsPerLine']
        #
        return theRet

    def _NormalizeData(self, data):
        """
        normalize to [0..1]
        """
        return (data - np.min(data)) / (np.max(data) - np.min(data))


def test_load_tif():
    path = "/Users/cudmore/data/dual-lcr/20210115/data/20210115__0001.tif"
    abf = bAbfText(path=path)

    import matplotlib.pyplot as plt

    plt.plot(abf.sweepX, abf.sweepY)
    plt.show()


if __name__ == "__main__":
    test_load_tif()
